"""
This module implements :class:`SpikeTrain`, an array of spike times.

:class:`SpikeTrain` derives from :class:`BaseNeo`, from
:module:`neo.core.baseneo`, and from :class:`quantities.Quantity`, which
inherits from :class:`numpy.array`.

Inheritance from :class:`numpy.array` is explained here:
http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

In brief:
* Initialization of a new object from constructor happens in :meth:`__new__`.
This is where user-specified attributes are set.

* :meth:`__array_finalize__` is called for all new objects, including those
created by slicing. This is where attributes are copied over from
the old object.
"""

from copy import deepcopy
import warnings

import numpy as np
import quantities as pq
from neo.core.baseneo import BaseNeo, MergeError, merge_annotations
from neo.core.dataobject import DataObject, ArrayDict

# need this to avoid circular import issue
import neo


def check_has_dimensions_time(*values) -> None:
    """
    Verify that all arguments have a dimensionality that is compatible
    with time.

    Raises
    ------
    ValueError if any value is not compatible with time

    Returns
    -------
    None: indicating that each value in values has the dimension time
    """
    errmsgs = []
    for value in values:
        dim = value.dimensionality.simplified
        if len(dim) != 1 or list(dim.values())[0] != 1 or not isinstance(list(dim.keys())[0], pq.UnitTime):
            errmsgs.append(f"value {value} has dimensions {dim}, not [time]")
    if errmsgs:
        raise ValueError("\n".join(errmsgs))


def _check_time_in_range(value, t_start, t_stop, view=False) -> None:
    """
    Verify that all times in `value` are between `t_start`
    and `t_stop` (inclusive)

    Parameters
    ----------
    value: array-like
        An array-like object with times
    t_start: float
        The starting time
    t_stop: float
        The stopping time
    view: bool, default: False
        If true views are used for the test. This increases speed but
        is only safe if certain that the dtype and units are the same.

    Raises
    ------
    ValueError
        * If t_stop < t_start
        * value.min() < t_start
        * value.max() > t_start

    Returns
    -------
    None: If check passes
    """

    if t_start > t_stop:
        raise ValueError(f"t_stop ({t_stop}) is before t_start ({t_start})")

    if not value.size:
        return

    if view:
        value = value.view(np.ndarray)
        t_start = t_start.view(np.ndarray)
        t_stop = t_stop.view(np.ndarray)

    if value.min() < t_start:
        raise ValueError(f"The first spike ({value}) is before t_start ({t_start})")
    if value.max() > t_stop:
        raise ValueError(f"The last spike ({value}) is after t_stop ({t_stop})")


def _check_waveform_dimensions(spiketrain) -> None:
    """
    Verify that waveform is compliant with the waveform definition as
    quantity array 3D (spike, channel, time)

    Parameters
    ----------
    spiketrain: neo.core.SpikeTrain
        The neo.core.SpikeTrain to check for waveforms

    Raises
    ------
    ValueError
        * number of waveforms doesn't match the number of spikes

    Returns
    -------
    None if test passes or no waveforms present
    """

    if not spiketrain.size:
        return

    waveforms = spiketrain.waveforms

    if (waveforms is None) or (not waveforms.size):
        return

    if waveforms.shape[0] != len(spiketrain):
        raise ValueError(
            f"Spiketrain length ({len(spiketrain)}) does not match to number of "
            f"waveforms present ({waveforms.shape[0]})"
        )


def _new_spiketrain(
    cls,
    signal,
    t_stop,
    units=None,
    dtype=None,
    copy=True,
    sampling_rate=1.0 * pq.Hz,
    t_start=0.0 * pq.s,
    waveforms=None,
    left_sweep=None,
    name=None,
    file_origin=None,
    description=None,
    array_annotations=None,
    annotations=None,
    segment=None,
    unit=None,
):
    """
    A function to map :meth:`BaseAnalogSignal.__new__` to function that
    does not do the unit checking. This is needed for :module:`pickle` to work.
    """
    if annotations is None:
        annotations = {}
    obj = SpikeTrain(
        signal,
        t_stop,
        units,
        dtype,
        copy,
        sampling_rate,
        t_start,
        waveforms,
        left_sweep,
        name,
        file_origin,
        description,
        array_annotations,
        **annotations,
    )
    obj.segment = segment
    obj.unit = unit
    return obj


def normalize_times_array(times, units=None, dtype=None, copy=True):
    """
    Return a quantity array with the correct units.
    There are four scenarios:

    A. times (NumPy array), units given as string or Quantities units
    B. times (Quantity array), units=None
    C. times (Quantity), units given as string or Quantities units
    D. times (NumPy array), units=None

    In scenarios A-C we return a tuple (times as a Quantity array, dimensionality)
    In scenario C, we rescale the original array to match `units`
    In scenario D, we raise a ValueError
    """
    if dtype is None:
        if not hasattr(times, "dtype"):
            dtype = float
    if units is None:
        # No keyword units, so get from `times`
        try:
            dim = times.units.dimensionality
        except AttributeError:
            raise ValueError("you must specify units")
    else:
        if hasattr(units, "dimensionality"):
            dim = units.dimensionality
        else:
            dim = pq.quantity.validate_dimensionality(units)

        if hasattr(times, "dimensionality"):
            if times.dimensionality.items() == dim.items():
                units = None  # units will be taken from times, avoids copying
            else:
                if not copy:
                    raise ValueError("cannot rescale and return view")
                else:
                    # this is needed because of a bug in python-quantities
                    # see issue # 65 in python-quantities github
                    # remove this if it is fixed
                    times = times.rescale(dim)

    # check to make sure the units are time
    # this approach is orders of magnitude faster than comparing the
    # reference dimensionality
    if len(dim) != 1 or list(dim.values())[0] != 1 or not isinstance(list(dim.keys())[0], pq.UnitTime):
        ValueError(f"Units have dimensions {dim.simplified}, not [time]")
    return pq.Quantity(times, units=units, dtype=dtype, copy=copy), dim


class SpikeTrain(DataObject):
    """
    :class:`SpikeTrain` is a :class:`Quantity` array of spike times.

    It is an ensemble of action potentials (spikes) emitted by the same unit
    in a period of time.

    Parameters
    ----------
    times: quantity array 1D | numpy array 1D | list
        The times of each spike.
    t_stop: quantity scalar | numpy scalar |float
        Time at which the SpikeTrain ended. This will be converted to thesame units as `times`.
        This argument is required because it specifies the period of time over which spikes could have occurred.
        Note that :attr:`t_start` is highly recommended for the same reason.
    units: (quantity units) | None, default: None
        Required if `times` is a list or numpy.ndarray`
        Not required if times is a quantities.Quantity
    dtype: numpy dtype | str | None, default: None
        Overrides the dtype of the times array if given.
        If None, the dtype of the times is used
    copy: bool, default: True
        Whether to copy the times array.
        Must be True when you request a change of units or dtype.
    sampling_rate: quantity scalar, default: 1.0 pq.Hz
        Number of samples per unit time for the waveforms.
    t_start: quantity scalar | numpy scalar | float
        Time at which the`SpikeTrain` began. This will be converted to the same units as `times`.
    waveforms: quantity array 3D (n_spikes, n_channels, n_time) | None, default: None
        The waveforms of each spike if given must be of the correct shape
         None indicates no waveforms
    left_sweep: (quantity array 1D) | None, default: None
        Time from the beginning of the waveform to the trigger time of the spike.
    name: str | None, default: None
        A label for the dataset.
    description: str | None, default: None
        A text description of this dataset
    file_origin: str | Filesystem path | URL | None, default: None
        The path or location of the original data file.
    array_annotations: dict
        A dictonary mapping of strings to numpy arrays containing annotations for all data points
    **annotations: dict
        Other user defined metadata given as a dict

    Notes
    -----
    Useful properties of a SpikeTrain

     * sampling_period: quantity scalar
            Interval between two samples (1/`sampling_rate`)
     * duration: quantity scalar
            Duration over which spikes can occur read-only (`t_stop` - `t_start`)
     * spike_duration: quantity scalar
            Duration of a waveform, read-only (`waveform`.shape[2] * `sampling_period`)
     * right_sweep: quantity scalar
            Time from the trigger times of the spikes to the end of the waveforms, read-only
            (`left_sweep` + `spike_duration`)
     * times: quantity array 1D
            Returns the :class:`SpikeTrain` as a quantity array.

    *Slicing*:
        :class:`SpikeTrain` objects can be sliced. When this occurs, a new
        :class:`SpikeTrain` (actually a view) is returned, with the same
        metadata, except that :attr:`waveforms` is also sliced in the same way
        (along dimension 0). Note that t_start and t_stop are not changed
        automatically, although you can still manually change them.

    Examples
    --------

    >>> from neo.core import SpikeTrain
    >>> from quantities import s
    >>>
    >>> train = SpikeTrain([3, 4, 5]*s, t_stop=10.0)
    >>> train2 = train[1:3]
    >>>
    >>> train.t_start
    array(0.0) * s
    >>> train.t_stop
    array(10.0) * s
    >>> train
    <SpikeTrain(array([ 3.,  4.,  5.]) * s, [0.0 s, 10.0 s])>
    >>> train2
    <SpikeTrain(array([ 4.,  5.]) * s, [0.0 s, 10.0 s])>

    """

    _parent_objects = ("Segment",)
    _parent_attrs = ("segment",)
    _quantity_attr = "times"
    _necessary_attrs = (("times", pq.Quantity, 1), ("t_start", pq.Quantity, 0), ("t_stop", pq.Quantity, 0))
    _recommended_attrs = (
        ("waveforms", pq.Quantity, 3),
        ("left_sweep", pq.Quantity, 0),
        ("sampling_rate", pq.Quantity, 0),
    ) + BaseNeo._recommended_attrs

    def __new__(
        cls,
        times,
        t_stop,
        units=None,
        dtype=None,
        copy=True,
        sampling_rate=1.0 * pq.Hz,
        t_start=0.0 * pq.s,
        waveforms=None,
        left_sweep=None,
        name=None,
        file_origin=None,
        description=None,
        array_annotations=None,
        **annotations,
    ):
        """
        Constructs a new :class:`Spiketrain` instance from data.

        This is called whenever a new :class:`SpikeTrain` is created from the
        constructor, but not when slicing.
        """
        if len(times) != 0 and waveforms is not None and len(times) != waveforms.shape[0]:
            # len(times)!=0 has been used to workaround a bug occurring during neo import
            raise ValueError("the number of waveforms should be equal to the number of spikes")

        if dtype is not None and hasattr(times, "dtype") and times.dtype != dtype:
            if not copy:
                raise ValueError("cannot change dtype and return view")

            # if t_start.dtype or t_stop.dtype != times.dtype != dtype,
            # _check_time_in_range can have problems, so we set the t_start
            # and t_stop dtypes to be the same as times before converting them
            # to dtype below
            # see ticket #38
            if hasattr(t_start, "dtype") and t_start.dtype != times.dtype:
                t_start = t_start.astype(times.dtype)
            if hasattr(t_stop, "dtype") and t_stop.dtype != times.dtype:
                t_stop = t_stop.astype(times.dtype)

        # Make sure units are consistent
        # also get the dimensionality now since it is much faster to feed
        # that to Quantity rather than a unit
        times, dim = normalize_times_array(times, units, dtype, copy)

        # Construct Quantity from data
        obj = times.view(cls)

        # spiketrain times always need to be 1-dimensional
        if len(obj.shape) > 1:
            raise ValueError("Spiketrain times array has more than 1 dimension")

        # if the dtype and units match, just copy the values here instead
        # of doing the much more expensive creation of a new Quantity
        # using items() is orders of magnitude faster
        if (
            hasattr(t_start, "dtype")
            and t_start.dtype == obj.dtype
            and hasattr(t_start, "dimensionality")
            and t_start.dimensionality.items() == dim.items()
        ):
            obj.t_start = t_start.copy()
        else:
            obj.t_start = pq.Quantity(t_start, units=dim, dtype=obj.dtype)

        if (
            hasattr(t_stop, "dtype")
            and t_stop.dtype == obj.dtype
            and hasattr(t_stop, "dimensionality")
            and t_stop.dimensionality.items() == dim.items()
        ):
            obj.t_stop = t_stop.copy()
        else:
            obj.t_stop = pq.Quantity(t_stop, units=dim, dtype=obj.dtype)

        # Store attributes
        obj.waveforms = waveforms
        obj.left_sweep = left_sweep
        obj.sampling_rate = sampling_rate

        # parents
        obj.segment = None
        obj.unit = None

        # Error checking (do earlier?)
        _check_time_in_range(obj, obj.t_start, obj.t_stop, view=True)

        return obj

    def __init__(
        self,
        times,
        t_stop,
        units=None,
        dtype=None,
        copy=True,
        sampling_rate=1.0 * pq.Hz,
        t_start=0.0 * pq.s,
        waveforms=None,
        left_sweep=None,
        name=None,
        file_origin=None,
        description=None,
        array_annotations=None,
        **annotations,
    ):
        """
        Initializes a newly constructed :class:`SpikeTrain` instance.
        """
        # This method is only called when constructing a new SpikeTrain,
        # not when slicing or viewing. We use the same call signature
        # as __new__ for documentation purposes. Anything not in the call
        # signature is stored in annotations.

        # Calls parent __init__, which grabs universally recommended
        # attributes and sets up self.annotations
        DataObject.__init__(
            self,
            name=name,
            file_origin=file_origin,
            description=description,
            array_annotations=array_annotations,
            **annotations,
        )

    def _repr_pretty_(self, pp, cycle):
        waveforms = ""
        if self.waveforms is not None:
            waveforms = " with waveforms"
        pp.text(
            f"{self.__class__.__name__} containing {self.size} spikes{waveforms}; "
            f"units {self.units.dimensionality.string}; datatype {self.dtype} "
        )
        if self._has_repr_pretty_attrs_():
            pp.breakable()
            self._repr_pretty_attrs_(pp, cycle)

        def _pp(line):
            pp.breakable()
            with pp.group(indent=1):
                pp.text(line)

        _pp(f"time: {self.t_start} to {self.t_stop}")

    def rescale(self, units, dtype=None):
        """
        Return a copy of the :class:`SpikeTrain` converted to the specified units
        The `dtype` argument exists only for backward compatibility within quantities, see
        https://github.com/python-quantities/python-quantities/pull/204
        """
        obj = super().rescale(units)
        obj.t_start = self.t_start.rescale(units)
        obj.t_stop = self.t_stop.rescale(units)
        obj.unit = self.unit
        return obj

    def __reduce__(self):
        """
        Map the __new__ function onto _new_BaseAnalogSignal, so that pickle
        works
        """
        import numpy

        return _new_spiketrain, (
            self.__class__,
            numpy.array(self),
            self.t_stop,
            self.units,
            self.dtype,
            True,
            self.sampling_rate,
            self.t_start,
            self.waveforms,
            self.left_sweep,
            self.name,
            self.file_origin,
            self.description,
            self.array_annotations,
            self.annotations,
            self.segment,
            self.unit,
        )

    def __array_finalize__(self, obj):
        """
        This is called every time a new :class:`SpikeTrain` is created.

        It is the appropriate place to set default values for attributes
        for :class:`SpikeTrain` constructed by slicing or viewing.

        User-specified values are only relevant for construction from
        constructor, and these are set in __new__. Then they are just
        copied over here.

        Note that the :attr:`waveforms` attribute is not sliced here. Nor is
        :attr:`t_start` or :attr:`t_stop` modified.
        """
        # This calls Quantity.__array_finalize__ which deals with
        # dimensionality
        super().__array_finalize__(obj)

        # Supposedly, during initialization from constructor, obj is supposed
        # to be None, but this never happens. It must be something to do
        # with inheritance from Quantity.
        if obj is None:
            return

        # Set all attributes of the new object `self` from the attributes
        # of `obj`. For instance, when slicing, we want to copy over the
        # attributes of the original object.
        self.t_start = getattr(obj, "t_start", None)
        self.t_stop = getattr(obj, "t_stop", None)
        self.waveforms = getattr(obj, "waveforms", None)
        self.left_sweep = getattr(obj, "left_sweep", None)
        self.sampling_rate = getattr(obj, "sampling_rate", None)
        self.segment = getattr(obj, "segment", None)
        self.unit = getattr(obj, "unit", None)

        # The additional arguments
        self.annotations = getattr(obj, "annotations", {})
        # Add empty array annotations, because they cannot always be copied,
        # but do not overwrite existing ones from slicing etc.
        # This ensures the attribute exists
        if not hasattr(self, "array_annotations"):
            self.array_annotations = ArrayDict(self._get_arr_ann_length())

        # Note: Array annotations have to be changed when slicing or initializing an object,
        # copying them over in spite of changed data would result in unexpected behaviour

        # Globally recommended attributes
        self.name = getattr(obj, "name", None)
        self.file_origin = getattr(obj, "file_origin", None)
        self.description = getattr(obj, "description", None)

        if hasattr(obj, "lazy_shape"):
            self.lazy_shape = obj.lazy_shape

    def __repr__(self):
        """
        Returns a string representing the :class:`SpikeTrain`.
        """
        return f"<SpikeTrain({super().__repr__()}, [{self.t_start}, {self.t_stop}])>"

    def sort(self):
        """
        Sorts the :class:`SpikeTrain` and its :attr:`waveforms`, if any,
        by time.
        """
        # sort the waveforms by the times
        sort_indices = np.argsort(self)
        if self.waveforms is not None and self.waveforms.any():
            self.waveforms = self.waveforms[sort_indices]
        self.array_annotate(**deepcopy(self.array_annotations_at_index(sort_indices)))

        # now sort the times
        # We have sorted twice, but `self = self[sort_indices]` introduces
        # a dependency on the slicing functionality of SpikeTrain.
        super().sort()

    def __getslice__(self, i, j):
        """
        Get a slice from :attr:`i` to :attr:`j`.

        Doesn't get called in Python 3, :meth:`__getitem__` is called instead
        """
        return self.__getitem__(slice(i, j))

    def __add__(self, time):
        """
        Shifts the time point of all spikes by adding the amount in
        :attr:`time` (:class:`Quantity`)

        If `time` is a scalar, this also shifts :attr:`t_start` and :attr:`t_stop`.
        If `time` is an array, :attr:`t_start` and :attr:`t_stop` are not changed unless
        some of the new spikes would be outside this range.
        In this case :attr:`t_start` and :attr:`t_stop` are modified if necessary to
        ensure they encompass all spikes.

        It is not possible to add two SpikeTrains (raises ValueError).
        """
        spikes = self.view(pq.Quantity)
        check_has_dimensions_time(time)
        if isinstance(time, SpikeTrain):
            raise TypeError("Can't add two spike trains")
        new_times = spikes + time
        if time.size > 1:
            t_start = min(self.t_start, np.min(new_times))
            t_stop = max(self.t_stop, np.max(new_times))
        else:
            t_start = self.t_start + time
            t_stop = self.t_stop + time
        return SpikeTrain(
            times=new_times,
            t_stop=t_stop,
            units=self.units,
            sampling_rate=self.sampling_rate,
            t_start=t_start,
            waveforms=self.waveforms,
            left_sweep=self.left_sweep,
            name=self.name,
            file_origin=self.file_origin,
            description=self.description,
            array_annotations=deepcopy(self.array_annotations),
            **self.annotations,
        )

    def __sub__(self, time):
        """
        Shifts the time point of all spikes by subtracting the amount in
        :attr:`time` (:class:`Quantity`)

        If `time` is a scalar, this also shifts :attr:`t_start` and :attr:`t_stop`.
        If `time` is an array, :attr:`t_start` and :attr:`t_stop` are not changed unless
        some of the new spikes would be outside this range.
        In this case :attr:`t_start` and :attr:`t_stop` are modified if necessary to
        ensure they encompass all spikes.

        In general, it is not possible to subtract two SpikeTrain objects (raises ValueError).
        However, if `time` is itself a SpikeTrain of the same size as the SpikeTrain,
        returns a Quantities array (since this is often used in checking
        whether two spike trains are the same or in calculating the inter-spike interval.
        """
        spikes = self.view(pq.Quantity)
        check_has_dimensions_time(time)
        if isinstance(time, SpikeTrain):
            if self.size == time.size:
                return spikes - time
            else:
                raise TypeError("Can't subtract spike trains with different sizes")
        else:
            new_times = spikes - time
            if time.size > 1:
                t_start = min(self.t_start, np.min(new_times))
                t_stop = max(self.t_stop, np.max(new_times))
            else:
                t_start = self.t_start - time
                t_stop = self.t_stop - time
            return SpikeTrain(
                times=spikes - time,
                t_stop=t_stop,
                units=self.units,
                sampling_rate=self.sampling_rate,
                t_start=t_start,
                waveforms=self.waveforms,
                left_sweep=self.left_sweep,
                name=self.name,
                file_origin=self.file_origin,
                description=self.description,
                array_annotations=deepcopy(self.array_annotations),
                **self.annotations,
            )

    def __getitem__(self, i):
        """
        Get the item or slice :attr:`i`.
        """
        obj = super().__getitem__(i)
        if hasattr(obj, "waveforms") and obj.waveforms is not None:
            obj.waveforms = obj.waveforms.__getitem__(i)
        try:
            obj.array_annotate(**deepcopy(self.array_annotations_at_index(i)))
        except AttributeError:  # If Quantity was returned, not SpikeTrain
            pass
        return obj

    def __setitem__(self, i, value):
        """
        Set the value the item or slice :attr:`i`.
        """
        if not hasattr(value, "units"):
            value = pq.Quantity(value, units=self.units)  # or should we be strict: raise ValueError(
            # "Setting a value  # requires a quantity")?
        # check for values outside t_start, t_stop
        _check_time_in_range(value, self.t_start, self.t_stop)
        super().__setitem__(i, value)

    def __setslice__(self, i, j, value):
        if not hasattr(value, "units"):
            value = pq.Quantity(value, units=self.units)
        _check_time_in_range(value, self.t_start, self.t_stop)
        super().__setslice__(i, j, value)

    def _copy_data_complement(self, other, deep_copy=False):
        """
        Copy the metadata from another :class:`SpikeTrain`.
        Note: Array annotations can not be copied here because length of data can change
        """
        # Note: Array annotations cannot be copied because length of data can be changed
        # here which would cause inconsistencies
        for attr in ("left_sweep", "sampling_rate", "name", "file_origin", "description", "annotations"):
            attr_value = getattr(other, attr, None)
            if deep_copy:
                attr_value = deepcopy(attr_value)
            setattr(self, attr, attr_value)

    def duplicate_with_new_data(self, signal, t_start=None, t_stop=None, waveforms=None, deep_copy=True, units=None):
        """
        Create a new :class:`SpikeTrain` with the same metadata
        but different data (times, t_start, t_stop)
        Note: Array annotations can not be copied here because length of data can change
        """
        # using previous t_start and t_stop if no values are provided
        if t_start is None:
            t_start = self.t_start
        if t_stop is None:
            t_stop = self.t_stop
        if waveforms is None:
            waveforms = self.waveforms
        if units is None:
            units = self.units
        else:
            units = pq.quantity.validate_dimensionality(units)

        new_st = self.__class__(signal, t_start=t_start, t_stop=t_stop, waveforms=waveforms, units=units)
        new_st._copy_data_complement(self, deep_copy=deep_copy)

        # Note: Array annotations are not copied here, because length of data could change

        # overwriting t_start and t_stop with new values
        new_st.t_start = t_start
        new_st.t_stop = t_stop

        # consistency check
        _check_time_in_range(new_st, new_st.t_start, new_st.t_stop, view=False)
        _check_waveform_dimensions(new_st)
        return new_st

    def time_slice(self, t_start, t_stop):
        """
        Creates a new :class:`SpikeTrain` corresponding to the time slice of
        the original :class:`SpikeTrain` between (and including) times
        :attr:`t_start` and :attr:`t_stop`. Either parameter can also be None
        to use infinite endpoints for the time interval.
        """
        _t_start = t_start
        _t_stop = t_stop
        if t_start is None:
            _t_start = -np.inf
        if t_stop is None:
            _t_stop = np.inf

        if _t_start > self.t_stop or _t_stop < self.t_start:
            # the alternative to raising an exception would be to return
            # a zero-duration spike train set at self.t_stop or self.t_start
            raise ValueError("A time slice completely outside the " "boundaries of the spike train is not defined.")

        indices = (self >= _t_start) & (self <= _t_stop)

        # Time slicing should create a deep copy of the object
        new_st = deepcopy(self[indices])

        new_st.t_start = max(_t_start, self.t_start)
        new_st.t_stop = min(_t_stop, self.t_stop)
        if self.waveforms is not None:
            new_st.waveforms = self.waveforms[indices]

        return new_st

    def time_shift(self, t_shift):
        """
        Shifts a :class:`SpikeTrain` to start at a new time.

        Parameters
        ----------
        t_shift: Quantity (time)
            Amount of time by which to shift the :class:`SpikeTrain`.

        Returns
        -------
        spiketrain: :class:`SpikeTrain`
            New instance of a :class:`SpikeTrain` object starting at t_shift later than the
            original :class:`SpikeTrain` (the original :class:`SpikeTrain` is not modified).
        """
        new_st = self.duplicate_with_new_data(
            signal=self.times.view(pq.Quantity) + t_shift, t_start=self.t_start + t_shift, t_stop=self.t_stop + t_shift
        )

        # Here we can safely copy the array annotations since we know that
        # the length of the SpikeTrain does not change.
        new_st.array_annotate(**self.array_annotations)

        return new_st

    def merge(self, *others):
        """
        Merge other :class:`SpikeTrain` objects into this one.

        The times of the :class:`SpikeTrain` objects combined in one array
        and sorted.

        If the attributes of the :class:`SpikeTrain` objects are not
        compatible, an Exception is raised.
        """
        for other in others:
            if isinstance(other, neo.io.proxyobjects.SpikeTrainProxy):
                raise MergeError(
                    "Cannot merge, SpikeTrainProxy objects cannot be merged"
                    "into regular SpikeTrain objects, please load them first."
                )
            elif not isinstance(other, SpikeTrain):
                raise MergeError("Cannot merge, only SpikeTrain" "can be merged into a SpikeTrain.")
            if self.sampling_rate != other.sampling_rate:
                raise MergeError("Cannot merge, different sampling rates")
            if self.t_start != other.t_start:
                raise MergeError("Cannot merge, different t_start")
            if self.t_stop != other.t_stop:
                raise MergeError("Cannot merge, different t_stop")
            if self.left_sweep != other.left_sweep:
                raise MergeError("Cannot merge, different left_sweep")
            if self.segment != other.segment:
                raise MergeError("Cannot merge these signals as they belong to" " different segments.")

        all_spiketrains = [self]
        all_spiketrains.extend([st.rescale(self.units) for st in others])

        wfs = [st.waveforms is not None for st in all_spiketrains]
        if any(wfs) and not all(wfs):
            raise MergeError("Cannot merge signal with waveform and signal " "without waveform.")
        stack = np.concatenate([np.asarray(st) for st in all_spiketrains])
        sorting = np.argsort(stack)
        stack = stack[sorting]

        kwargs = {}

        kwargs["array_annotations"] = self._merge_array_annotations(others, sorting=sorting)

        for name in ("name", "description", "file_origin"):
            attr = getattr(self, name)

            # check if self is already a merged spiketrain
            # if it is, get rid of the bracket at the end to append more attributes
            if attr is not None:
                if attr.startswith("merge(") and attr.endswith(")"):
                    attr = attr[:-1]

            for other in others:
                attr_other = getattr(other, name)

                # both attributes are None --> nothing to do
                if attr is None and attr_other is None:
                    continue

                # one of the attributes is None --> convert to string in order to merge them
                elif attr is None or attr_other is None:
                    attr = str(attr)
                    attr_other = str(attr_other)

                # check if the other spiketrain is already a merged spiketrain
                # if it is, append all of its merged attributes that aren't already in attr
                if attr_other.startswith("merge(") and attr_other.endswith(")"):
                    for subattr in attr_other[6:-1].split("; "):
                        if subattr not in attr:
                            attr += "; " + subattr
                            if not attr.startswith("merge("):
                                attr = "merge(" + attr

                # if the other attribute is not in the list --> append
                # if attr doesn't already start with merge add merge( in the beginning
                elif attr_other not in attr:
                    attr += "; " + attr_other
                    if not attr.startswith("merge("):
                        attr = "merge(" + attr

            # close the bracket of merge(...) if necessary
            if attr is not None:
                if attr.startswith("merge("):
                    attr += ")"

            # write attr into kwargs dict
            kwargs[name] = attr

        merged_annotations = merge_annotations(*(st.annotations for st in all_spiketrains))
        kwargs.update(merged_annotations)

        train = SpikeTrain(
            stack,
            units=self.units,
            dtype=self.dtype,
            copy=False,
            t_start=self.t_start,
            t_stop=self.t_stop,
            sampling_rate=self.sampling_rate,
            left_sweep=self.left_sweep,
            **kwargs,
        )
        if all(wfs):
            wfs_stack = np.vstack([st.waveforms.rescale(self.waveforms.units) for st in all_spiketrains])
            wfs_stack = wfs_stack[sorting] * self.waveforms.units
            train.waveforms = wfs_stack
        train.segment = self.segment
        if train.segment is not None:
            self.segment.spiketrains.append(train)

        return train

    def _merge_array_annotations(self, others, sorting=None):
        """
        Merges array annotations of multiple different objects.
        The merge happens in such a way that the result fits the merged data
        In general this means concatenating the arrays from the objects.
        If an annotation is not present in one of the objects, it will be omitted.
        Apart from that the array_annotations need to be sorted according to the sorting of
        the spikes.
        :return Merged array_annotations
        """
        if sorting is None:
            raise ValueError("The order of the merged spikes must be known")

        merged_array_annotations = {}

        omitted_keys_self = []

        keys = self.array_annotations.keys()
        for key in keys:
            try:
                self_ann = deepcopy(self.array_annotations[key])
                other_ann = np.concatenate([deepcopy(other.array_annotations[key]) for other in others])
                if isinstance(self_ann, pq.Quantity):
                    other_ann.rescale(self_ann.units)
                    arr_ann = np.concatenate([self_ann, other_ann]) * self_ann.units
                else:
                    arr_ann = np.concatenate([self_ann, other_ann])
                merged_array_annotations[key] = arr_ann[sorting]
            # Annotation only available in 'self', must be skipped
            # Ignore annotations present only in one of the SpikeTrains
            except KeyError:
                omitted_keys_self.append(key)
                continue

        omitted_keys_other = [
            key
            for key in np.unique([key for other in others for key in other.array_annotations])
            if key not in self.array_annotations
        ]

        if omitted_keys_self or omitted_keys_other:
            warnings.warn(
                f"The following array annotations were omitted, because they were only "
                f"present in one of the merged objects: {omitted_keys_self} from the "
                f"one that was merged into and {omitted_keys_other} from the ones that "
                f"were merged into it.",
                UserWarning,
            )

        return merged_array_annotations

    @property
    def times(self):
        """
        Returns the :class:`SpikeTrain` as a quantity array.
        """
        return pq.Quantity(self)

    @property
    def duration(self):
        """
        Duration over which spikes can occur,

        (:attr:`t_stop` - :attr:`t_start`)
        """
        if self.t_stop is None or self.t_start is None:
            return None
        return self.t_stop - self.t_start

    @property
    def spike_duration(self):
        """
        Duration of a waveform.

        (:attr:`waveform`.shape[2] * :attr:`sampling_period`)
        """
        if self.waveforms is None or self.sampling_rate is None:
            return None
        return self.waveforms.shape[2] / self.sampling_rate

    @property
    def sampling_period(self):
        """
        Interval between two samples.

        (1/:attr:`sampling_rate`)
        """
        if self.sampling_rate is None:
            return None
        return 1.0 / self.sampling_rate

    @sampling_period.setter
    def sampling_period(self, period):
        """
        Setter for :attr:`sampling_period`
        """
        if period is None:
            self.sampling_rate = None
        else:
            self.sampling_rate = 1.0 / period

    @property
    def right_sweep(self):
        """
        Time from the trigger times of the spikes to the end of the waveforms.

        (:attr:`left_sweep` + :attr:`spike_duration`)
        """
        dur = self.spike_duration
        if self.left_sweep is None or dur is None:
            return None
        return self.left_sweep + dur



"""
BCI2000RawIO is a class to read BCI2000 .dat files.
https://www.bci2000.org/mediawiki/index.php/Technical_Reference:BCI2000_File_Format
"""

import numpy as np
import re

try:
    from urllib.parse import unquote
except ImportError:
    from urllib import url2pathname as unquote

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from neo.core import NeoReadWriteError


class BCI2000RawIO(BaseRawIO):
    """
    Class for reading data from a BCI2000 .dat file, either version 1.0 or 1.1

    Parameters
    ----------
    filename: str, default: ''
        The *.dat recorded by BCI2000
    """

    extensions = ["dat"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        BaseRawIO.__init__(self)
        self.filename = filename
        self._my_events = None

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        file_info, state_defs, param_defs = parse_bci2000_header(self.filename)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]

        # one unique stream
        signal_streams = np.array([("Signals", "0")], dtype=_signal_stream_dtype)
        self.header["signal_streams"] = signal_streams

        sig_channels = []
        for chan_ix in range(file_info["SourceCh"]):
            if "ChannelNames" in param_defs and isinstance(param_defs["ChannelNames"]["value"], (list, tuple)):
                ch_name = param_defs["ChannelNames"]["value"][chan_ix]
            else:
                ch_name = "ch" + str(chan_ix)
            chan_id = str(chan_ix + 1)
            sr = param_defs["SamplingRate"]["value"]  # Hz
            dtype = file_info["DataFormat"]
            units = "uV"
            gain = param_defs["SourceChGain"]["value"][chan_ix]

            if isinstance(gain, str):
                r = re.findall(r"(\d+)(\D+)", gain)
                # some files have strange units attached to gain
                # in that case it is ignored
                if len(r) == 1:
                    gain = r[0][0]
                gain = float(gain)

            offset = param_defs["SourceChOffset"]["value"][chan_ix]
            if isinstance(offset, str):
                offset = float(offset)

            stream_id = "0"
            sig_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))
        self.header["signal_channels"] = np.array(sig_channels, dtype=_signal_channel_dtype)

        self.header["spike_channels"] = np.array([], dtype=_spike_channel_dtype)

        # creating event channel for each state variable
        event_channels = []
        for st_ix, st_tup in enumerate(state_defs):
            event_channels.append((st_tup[0], "ev_" + str(st_ix), "event"))
        self.header["event_channels"] = np.array(event_channels, dtype=_event_channel_dtype)

        # Add annotations.

        # Generates basic annotations in nested dict self.raw_annotations
        self._generate_minimal_annotations()

        self.raw_annotations["blocks"][0].update({"file_info": file_info, "param_defs": param_defs})
        event_annotations = self.raw_annotations["blocks"][0]["segments"][0]["events"]
        for ev_ix, ev_dict in enumerate(event_annotations):
            ev_dict.update(
                {
                    "length": state_defs[ev_ix][1],
                    "startVal": state_defs[ev_ix][2],
                    "bytePos": state_defs[ev_ix][3],
                    "bitPos": state_defs[ev_ix][4],
                }
            )

        import time

        time_formats = ["%a %b %d %H:%M:%S %Y", "%Y-%m-%dT%H:%M:%S"]
        try:
            self._global_time = time.mktime(time.strptime(param_defs["StorageTime"]["value"], time_formats[0]))
        except:
            self._global_time = time.mktime(time.strptime(param_defs["StorageTime"]["value"], time_formats[1]))

        # Save variables to make it easier to load the binary data.
        self._read_info = {
            "header_len": file_info["HeaderLen"],
            "n_chans": file_info["SourceCh"],
            "sample_dtype": {"int16": np.int16, "int32": np.int32, "float32": np.float32}.get(file_info["DataFormat"]),
            "state_vec_len": file_info["StatevectorLen"],
            "sampling_rate": param_defs["SamplingRate"]["value"],
        }
        # Calculate the dtype for a single timestamp of data. This contains the data + statevector
        self._read_info["line_dtype"] = [
            ("raw_vector", self._read_info["sample_dtype"], self._read_info["n_chans"]),
            ("state_vector", np.uint8, self._read_info["state_vec_len"]),
        ]
        import os

        self._read_info["n_samps"] = int(
            (os.stat(self.filename).st_size - file_info["HeaderLen"]) / np.dtype(self._read_info["line_dtype"]).itemsize
        )

        # memmap is fast so we can get the data ready for reading now.
        self._memmap = np.memmap(
            self.filename, dtype=self._read_info["line_dtype"], offset=self._read_info["header_len"], mode="r"
        )

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        return self._read_info["n_samps"] / self._read_info["sampling_rate"]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._read_info["n_samps"]

    def _get_signal_t_start(self, block_index, seg_index, channel_indexes):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._read_info["n_samps"]

        if (i_start < 0) or (i_start > self._read_info["n_samps"]):
            raise ValueError("`i_start` is outside of data range")
        if (i_stop < 0) or (i_stop > self._read_info["n_samps"]):
            raise ValueError("`i_stop` is outside of data range")

        if channel_indexes is None:
            channel_indexes = np.arange(self.header["signal_channels"].size)
        return self._memmap["raw_vector"][i_start:i_stop, channel_indexes]

    def _spike_count(self, block_index, seg_index, unit_index):
        return 0

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        return None

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        return None

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        return None

    def _event_count(self, block_index, seg_index, event_channel_index):
        return self._event_arrays_list[event_channel_index][0].shape[0]

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        # Return 3 numpy arrays: timestamp, durations, labels
        # durations must be None for 'event'
        # label must a dtype ='U'
        ts, dur, labels = self._event_arrays_list[event_channel_index]
        # seg_t_start = self._segment_t_start(block_index, seg_index)
        keep = np.ones(ts.shape, dtype=bool)
        if t_start is not None:
            keep = np.logical_and(keep, ts >= t_start)
        if t_stop is not None:
            keep = np.logical_and(keep, ts <= t_stop)
        return ts[keep], dur[keep], labels[keep]

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = (event_timestamps / float(self._read_info["sampling_rate"])).astype(dtype)
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        durations = (raw_duration / float(self._read_info["sampling_rate"])).astype(dtype)
        return durations

    @property
    def _event_arrays_list(self):
        if self._my_events is None:
            event_annotations = self.raw_annotations["blocks"][0]["segments"][0]["events"]

            self._my_events = []
            for event_channel_index in range(self.event_channels_count()):
                sd = event_annotations[event_channel_index]
                ev_times = durs = vals = np.array([])
                # Skip these big but mostly useless (?) states.
                if sd["name"] not in ["SourceTime", "StimulusTime"]:
                    # Determine which bytes of self._memmap['state_vector'] are needed.
                    nbytes = int(np.ceil((sd["bitPos"] + sd["length"]) / 8))
                    byte_slice = slice(sd["bytePos"], sd["bytePos"] + nbytes)
                    # Then determine how to mask those bytes to get only the needed bits.
                    bit_mask = np.array([255] * nbytes, dtype=np.uint8)
                    bit_mask[0] &= 255 & (255 << sd["bitPos"])  # Fix the mask for the first byte
                    extra_bits = 8 - (sd["bitPos"] + sd["length"]) % 8
                    bit_mask[-1] &= 255 & (255 >> extra_bits)  # Fix the mask for the last byte
                    # When converting to an int, we need to know which integer type it will become
                    n_max_bytes = 1 << (nbytes - 1).bit_length()
                    view_type = {1: np.int8, 2: np.int16, 4: np.int32, 8: np.int64}.get(n_max_bytes)
                    # Slice and mask the data
                    masked_byte_array = self._memmap["state_vector"][:, byte_slice] & bit_mask
                    # Convert byte array to a vector of ints:
                    # pad to give even columns then view as larger int type
                    state_vec = np.pad(masked_byte_array, (0, n_max_bytes - nbytes), "constant").view(dtype=view_type)
                    state_vec = np.right_shift(state_vec, sd["bitPos"])[:, 0]

                    # In the state vector, find 'events' whenever the state changes
                    st_ch_ix = np.where(np.hstack((0, np.diff(state_vec))) != 0)[0]  # event inds
                    if len(st_ch_ix) > 0:
                        ev_times = st_ch_ix
                        durs = np.asarray([None] * len(st_ch_ix))
                        # np.hstack((np.diff(st_ch_ix), len(state_vec) - st_ch_ix[-1]))
                        vals = np.char.mod("%d", state_vec[st_ch_ix])  # event val, string'd

                self._my_events.append([ev_times, durs, vals.astype("U")])

        return self._my_events


def parse_bci2000_header(filename):
    # typically we want parameter values in Hz, seconds, or microvolts.
    scales_dict = {
        "hz": 1,
        "khz": 1000,
        "mhz": 1000000,
        "uv": 1,
        "muv": 1,
        "mv": 1000,
        "v": 1000000,
        "s": 1,
        "us": 0.000001,
        "mus": 0.000001,
        "ms": 0.001,
        "min": 60,
        "sec": 1,
        "usec": 0.000001,
        "musec": 0.000001,
        "msec": 0.001,
    }

    def rescale_value(param_val, data_type):
        unit_str = ""
        if param_val.lower().startswith("0x"):
            param_val = int(param_val, 16)
        elif data_type in ["int", "float"]:
            matches = re.match(r"(-*\d+)(\w*)", param_val)
            if matches is not None:  # Can be None for % in def, min, max vals
                param_val, unit_str = matches.group(1), matches.group(2)
                param_val = int(param_val) if data_type == "int" else float(param_val)
                if len(unit_str) > 0:
                    param_val *= scales_dict.get(unit_str.lower(), 1)
        else:
            param_val = unquote(param_val)
        return param_val, unit_str

    def parse_dimensions(param_list):
        num_els = param_list.pop(0)
        # Sometimes the number of elements isn't given,
        # but the list of element labels is wrapped with {}
        if num_els == "{":
            num_els = param_list.index("}")
            el_labels = [unquote(param_list.pop(0)) for x in range(num_els)]
            param_list.pop(0)  # Remove the '}'
        else:
            num_els = int(num_els)
            el_labels = [str(ix) for ix in range(num_els)]
        return num_els, el_labels

    with open(filename, "rb") as fid:

        # Parse the file header (plain text)

        # The first line contains basic information which we store in a dictionary.
        temp = fid.readline().decode("utf8").split()
        keys = [k.rstrip("=") for k in temp[::2]]
        vals = temp[1::2]
        # Insert default version and format
        file_info = {"BCI2000V": 1.0, "DataFormat": "int16"}
        file_info.update(**dict(zip(keys, vals)))
        # From string to float/int
        file_info["BCI2000V"] = float(file_info["BCI2000V"])
        for k in ["HeaderLen", "SourceCh", "StatevectorLen"]:
            if k in file_info:
                file_info[k] = int(file_info[k])

        # The next lines contain state vector definitions.
        temp = fid.readline().decode("utf8").strip()
        if temp != "[ State Vector Definition ]":
            raise NeoReadWriteError(f"State definitions not found in header {filename}")
        state_defs = []
        state_def_dtype = [("name", "a64"), ("length", int), ("startVal", int), ("bytePos", int), ("bitPos", int)]
        while True:
            temp = fid.readline().decode("utf8").strip()
            if len(temp) == 0 or temp[0] == "[":
                # Presence of '[' signifies new section.
                break
            temp = temp.split()
            state_defs.append((temp[0], int(temp[1]), int(temp[2]), int(temp[3]), int(temp[4])))
        state_defs = np.array(state_defs, dtype=state_def_dtype)

        # The next lines contain parameter definitions.
        # There are many, and their formatting can be complicated.
        if temp != "[ Parameter Definition ]":
            raise NeoReadWriteError(f"Parameter definitions not found in header {filename}")
        param_defs = {}
        while True:
            temp = fid.readline().decode("utf8")
            if fid.tell() >= file_info["HeaderLen"]:
                # End of header.
                break
            if len(temp.strip()) == 0:
                continue  # Skip empty lines
            # Everything after the '//' is a comment.
            temp = temp.strip().split("//", 1)
            param_def = {"comment": temp[1].strip() if len(temp) > 1 else ""}
            # Parse the parameter definition. Generally it is sec:cat:name dtype name param_value+
            temp = temp[0].split()
            param_def.update({"section_category_name": [unquote(x) for x in temp.pop(0).split(":")]})
            dtype = temp.pop(0)
            param_name = unquote(temp.pop(0).rstrip("="))
            # Parse the rest. Parse method depends on the dtype
            param_value, units = None, None
            if dtype in ("int", "float"):
                param_value = temp.pop(0)
                if param_value == "auto":
                    param_value = np.nan
                    units = ""
                else:
                    param_value, units = rescale_value(param_value, dtype)
            elif dtype in ("string", "variant"):
                param_value = unquote(temp.pop(0))
            elif dtype.endswith("list"):  # e.g., intlist, stringlist, floatlist, list
                dtype = dtype[:-4]
                # The list parameter values will begin with either
                # an int to specify the number of elements
                # or a list of labels surrounded by { }.
                num_elements, element_labels = parse_dimensions(temp)  # This will pop off info.
                param_def.update({"element_labels": element_labels})
                pv_un = [rescale_value(pv, dtype) for pv in temp[:num_elements]]
                if len(pv_un) > 0:
                    param_value, units = zip(*pv_un)
                else:
                    param_value, units = np.nan, ""
                temp = temp[num_elements:]
                # Sometimes an element list will be a list of ints even though
                # the element_type is '' (str)...
                # This usually happens for known parameters, such as SourceChOffset,
                # that can be dealt with explicitly later.
            elif dtype.endswith("matrix"):
                dtype = dtype[:-6]
                # The parameter values will be preceded by two dimension descriptors,
                # first rows then columns. Each dimension might be described by an
                # int or a list of labels surrounded by {}
                n_rows, row_labels = parse_dimensions(temp)
                n_cols, col_labels = parse_dimensions(temp)
                param_def.update({"row_labels": row_labels, "col_labels": col_labels})

                param_value = []
                units = []
                for row_ix in range(n_rows):
                    cols = []
                    for col_ix in range(n_cols):
                        col_val, _units = rescale_value(temp[row_ix * n_cols + col_ix], dtype)
                        cols.append(col_val)
                        units.append(_units)
                    param_value.append(cols)
                temp = temp[n_rows * n_cols :]

            param_def.update({"value": param_value, "units": units, "dtype": dtype})

            # At the end of the parameter definition, we might get
            # default, min, max values for the parameter.
            temp.reverse()
            if len(temp):
                param_def.update({"max_val": rescale_value(temp.pop(0), dtype)})
            if len(temp):
                param_def.update({"min_val": rescale_value(temp.pop(0), dtype)})
            if len(temp):
                param_def.update({"default_val": rescale_value(temp.pop(0), dtype)})

            param_defs.update({param_name: param_def})
            # End parameter block
    # Outdent to close file
    return file_info, state_defs, param_defs



"""
This module implement the "old" OpenEphys format.
In this format channels are split into several files

https://open-ephys.github.io/gui-docs/User-Manual/Recording-data/Open-Ephys-format.html


Author: Samuel Garcia
"""

import re
from pathlib import Path

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from neo.core import NeoReadWriteError


RECORD_SIZE = 1024
HEADER_SIZE = 1024


class OpenEphysRawIO(BaseRawIO):
    """
    Class for reading openephys data

    Parameters
    ----------
    dirname: str
        The directory where the files are stored.
    ignore_timestamps_errors: bool
        (deprecated) This parameter is not used anymore.
    fill_gap_value: int
        When gaps are detected in continuous files, the gap is filled with this value.
        Default is 0.

    Notes
    -----
    OpenEphys GUI software offers several data formats, see
    https://open-ephys.atlassian.net/wiki/spaces/OEW/pages/491632/Data+format

    This class implements the legacy OpenEphys format here
    https://open-ephys.atlassian.net/wiki/spaces/OEW/pages/65667092/Open+Ephys+format

    The OpenEphys group already proposes some tools here:
    https://github.com/open-ephys/analysis-tools/blob/master/OpenEphys.py
    but (i) there is no package at PyPI and (ii) those tools read everything in memory.

    The format is directory based with several files:
        * .continuous
        * .events
        * .spikes

    This implementation is based on:
      * this code https://github.com/open-ephys/analysis-tools/blob/master/Python3/OpenEphys.py
        written by Dan Denman and Josh Siegle
      * a previous PR by Cristian Tatarau at Charité Berlin

    In contrast to previous code for reading this format, here all data use memmap so it should
    be super fast and light compared to legacy code.

    When the acquisition is stopped and restarted then files are named ``*_2``, ``*_3``.
    In that case this class creates a new Segment. Note that timestamps are reset in this
    situation.

    Limitation :
      * Works only if all continuous channels have the same sampling rate, which is a reasonable
        hypothesis.
      * A recording can contain gaps due to USB stream loss when high CPU load when recording.
        These gaps are checked channel per channel which makes the parse_header() slow.
        If gaps are detected then they are filled with zeros but then the reading will be much slower for getting signals.

    """

    # file formats used by openephys
    extensions = ["continuous", "openephys", "spikes", "events", "xml"]
    rawmode = "one-dir"

    def __init__(self, dirname="", ignore_timestamps_errors=None, fill_gap_value=0):
        BaseRawIO.__init__(self)
        self.dirname = dirname
        self.fill_gap_value = int(fill_gap_value)
        if ignore_timestamps_errors is not None:
            self.logger.warning("OpenEphysRawIO ignore_timestamps_errors=True/False is not used anymore")

    def _source_name(self):
        return self.dirname

    def _parse_header(self):
        info = self._info = explore_folder(self.dirname)
        nb_segment = info["nb_segment"]

        # scan for continuous files
        self._sigs_memmap = {}
        self._sig_length = {}
        self._sig_timestamp0 = {}
        self._sig_has_gap = {}
        self._gap_mode = False
        signal_channels = []
        oe_indices = sorted(list(info["continuous"].keys()))
        for seg_index, oe_index in enumerate(oe_indices):
            self._sigs_memmap[seg_index] = {}
            self._sig_has_gap[seg_index] = {}

            all_first_timestamps = []
            all_last_timestamps = []
            all_samplerate = []
            for chan_index, continuous_filename in enumerate(info["continuous"][oe_index]):
                chan_info = read_file_header(continuous_filename)

                s = continuous_filename.stem.split("_")
                # Formats are ['processor_id', 'ch_name'] or  ['processor_id', 'name', 'ch_name']
                if len(s) == 2:
                    processor_id, ch_name = s[0], s[1]
                    chan_str = re.split(r"(\d+)", s[1])[0]
                else:
                    processor_id, ch_name = s[0], s[2]
                    chan_str = re.split(r"(\d+)", s[2])[0]
                # note that chan_id is not unique in case of CH + AUX
                chan_id = int(ch_name.replace(chan_str, ""))

                filesize = continuous_filename.stat().st_size
                size = (filesize - HEADER_SIZE) // np.dtype(continuous_dtype).itemsize
                data_chan = np.memmap(
                    continuous_filename, mode="r", offset=HEADER_SIZE, dtype=continuous_dtype, shape=(size,)
                )
                self._sigs_memmap[seg_index][chan_index] = data_chan

                all_first_timestamps.append(data_chan[0]["timestamp"])
                all_last_timestamps.append(data_chan[-1]["timestamp"] + RECORD_SIZE)
                all_samplerate.append(chan_info["sampleRate"])

                # check for continuity (no gaps)
                diff = np.diff(data_chan["timestamp"])
                channel_has_gaps = not np.all(diff == RECORD_SIZE)
                self._sig_has_gap[seg_index][chan_index] = channel_has_gaps

                if channel_has_gaps:
                    # protect against strange timestamp block like in file 'OpenEphys_SampleData_3' CH32
                    if not np.median(diff) == RECORD_SIZE:
                        raise NeoReadWriteError(
                            f"This file has a non valid data block size for channel {chan_id}, this case cannot be handled"
                        )

                channel_has_corrupted_timestamps = np.any(diff <= 0)
                if channel_has_corrupted_timestamps:
                    # protect against corrupted timestamp in channel
                    raise ValueError(
                        f"{ch_name} has timestamps with zero values or negative differences between consecutive timestamps, this file ({continuous_filename}) with corrupted timestamps needs to be moved away from the folder."
                    )

                if seg_index == 0:
                    # add in channel list
                    if ch_name[:2].upper() == "CH":
                        units = "uV"
                    else:
                        units = "V"
                    signal_channels.append(
                        (
                            ch_name,
                            chan_id,
                            chan_info["sampleRate"],
                            "int16",
                            units,
                            chan_info["bitVolts"],
                            0.0,
                            processor_id,
                        )
                    )

            if any(self._sig_has_gap[seg_index].values()):
                channel_with_gaps = list(self._sig_has_gap[seg_index].keys())
                self.logger.warning(
                    f"This OpenEphys dataset contains gaps for some channels {channel_with_gaps} in segment {seg_index} the read will be slow"
                )
                self._gap_mode = True

            if not all(all_first_timestamps[0] == e for e in all_first_timestamps) or not all(
                all_last_timestamps[0] == e for e in all_last_timestamps
            ):
                # In some cases, continuous do not have the same length because
                # we need to clip
                self.logger.warning(
                    "Continuous files do not have aligned timestamps; " "clipping to make them aligned."
                )

                first = max(all_first_timestamps)
                last = min(all_last_timestamps)
                for chan_index in self._sigs_memmap[seg_index]:
                    data_chan = self._sigs_memmap[seg_index][chan_index]
                    keep = (data_chan["timestamp"] >= first) & (data_chan["timestamp"] < last)
                    data_chan = data_chan[keep]
                    self._sigs_memmap[seg_index][chan_index] = data_chan
            else:
                # no clip
                first = all_first_timestamps[0]
                last = all_last_timestamps[0]

            # check unique sampling rate
            if not all(all_samplerate[0] == e for e in all_samplerate):
                raise NeoReadWriteError("Not all signals have the same sample rate")

            self._sig_length[seg_index] = last - first
            self._sig_timestamp0[seg_index] = first

        if len(signal_channels) > 0:
            signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)
            self._sig_sampling_rate = signal_channels["sampling_rate"][0]  # unique for channel

            # split channels in stream depending the name CHxxx ADCxxx
            chan_stream_ids = [name[:2] if name.startswith("CH") else name[:3] for name in signal_channels["name"]]
            signal_channels["stream_id"] = chan_stream_ids

            # and create streams channels (keep natural order 'CH' first)
            stream_ids, order = np.unique(chan_stream_ids, return_index=True)
            stream_ids = stream_ids[np.argsort(order)]
            signal_streams = [(f"Signals {stream_id}", f"{stream_id}") for stream_id in stream_ids]
            signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)
        else:
            signal_streams = np.array([])
        # scan for spikes files
        spike_channels = []

        if len(info["spikes"]) > 0:
            self._first_spk_timestamps = []
            self._last_spk_timestamps = []
            self._spikes_memmap = {}
            oe_indices_spk = sorted(list(info["spikes"].keys()))
            for seg_index, oe_index in enumerate(oe_indices_spk):
                self._spikes_memmap[seg_index] = {}
                for spike_filename in info["spikes"][oe_index]:
                    spike_info = read_file_header(spike_filename)
                    spikes_dtype = make_spikes_dtype(spike_filename)

                    # "STp106.0n0_2.spikes" to "STp106.0n0"
                    name = spike_filename.stem
                    if seg_index > 0:
                        name = name.replace("_" + str(seg_index + 1), "")

                    data_spike = np.memmap(spike_filename, mode="r", offset=HEADER_SIZE, dtype=spikes_dtype)
                    self._spikes_memmap[seg_index][name] = data_spike

                    self._first_spk_timestamps.append(data_spike[0]["timestamp"])
                    self._last_spk_timestamps.append(data_spike[-1]["timestamp"])

            # In each file 'sorted_id' indicate the number of cluster so number of units
            # so need to scan file for all segment to get units
            self._spike_sampling_rate = None
            for spike_filename_seg0 in info["spikes"][0]:
                name = spike_filename_seg0.stem

                spike_info = read_file_header(spike_filename_seg0)
                if self._spike_sampling_rate is None:
                    self._spike_sampling_rate = spike_info["sampleRate"]
                else:
                    if self._spike_sampling_rate != spike_info["sampleRate"]:
                        raise ValueError("There is a mismatch in spike sampling rate")

                # scan all to detect several all unique(sorted_ids)
                all_sorted_ids = []
                for seg_index in range(nb_segment):
                    data_spike = self._spikes_memmap[seg_index][name]
                    all_sorted_ids += np.unique(data_spike["sorted_id"]).tolist()
                all_sorted_ids = np.unique(all_sorted_ids)

                # suppose all channels have the same gain
                wf_units = "uV"
                wf_gain = 1000.0 / data_spike[0]["gains"][0]
                wf_offset = -(2**15) * wf_gain
                wf_left_sweep = 0
                wf_sampling_rate = spike_info["sampleRate"]

                # each sorted_id is one channel
                for sorted_id in all_sorted_ids:
                    unit_name = f"{name}#{sorted_id}"
                    unit_id = f"{name}#{sorted_id}"
                    spike_channels.append(
                        (unit_name, unit_id, wf_units, wf_gain, wf_offset, wf_left_sweep, wf_sampling_rate)
                    )

        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # event file are:
        #    * all_channel.events (header + binary)  -->  event 0
        #    * n_RhythmData-a.events (header + binary) --> event 0 (maybe a new naming convention? )
        # and message.events (text based)      --> event 1 not implemented yet
        event_channels = []
        self._events_memmap = {}
        event_files = list(
            [
                event_file
                for event_file in Path(self.dirname).glob("**/*.events")
                if event_file.name != "messages.events"
            ]
        )
        event_files.sort()  # sort should put the xx.events first followed by xx_x.events
        # only run if we have actual potential event files
        if len(event_files) > 0:
            event_file_name_0 = event_files[0].stem  # this should always be the file without a '_n' appended
            for seg_index, oe_index in enumerate(oe_indices):
                if oe_index == 0:
                    event_filename = Path(self.dirname) / (event_file_name_0 + ".events")
                else:
                    event_filename = Path(self.dirname) / (event_file_name_0 + f"_{oe_index + 1}.events")

                event_info = read_file_header(event_filename)
                # event files can exist, but just not have data
                try:
                    self._event_sampling_rate = event_info["sampleRate"]
                except KeyError:
                    break
                data_event = np.memmap(event_filename, mode="r", offset=HEADER_SIZE, dtype=events_dtype)
                self._events_memmap[seg_index] = data_event
        # only append event channels if they actually exist & have data
        if len(self._events_memmap.keys()) > 0:
            event_channels.append((event_filename.stem, "", "event"))
        # event_channels.append(('message', '', 'event')) # not implemented
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # main header
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [nb_segment]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Annotate some objects from continuous files
        self._generate_minimal_annotations()
        bl_ann = self.raw_annotations["blocks"][0]
        for seg_index, oe_index in enumerate(oe_indices):
            seg_ann = bl_ann["segments"][seg_index]
            if len(info["continuous"]) > 0:
                fullname = info["continuous"][oe_index][0]
                chan_info = read_file_header(fullname)
                seg_ann["openephys_version"] = chan_info["version"]
                bl_ann["openephys_version"] = chan_info["version"]
                seg_ann["date_created"] = chan_info["date_created"]
                seg_ann["openephys_segment_index"] = oe_index + 1

    def _segment_t_start(self, block_index, seg_index):
        # segment start/stop are defined by continuous channels
        return self._sig_timestamp0[seg_index] / self._sig_sampling_rate

    def _segment_t_stop(self, block_index, seg_index):
        return (self._sig_timestamp0[seg_index] + self._sig_length[seg_index]) / self._sig_sampling_rate

    def _get_signal_size(self, block_index, seg_index, stream_index):
        return self._sig_length[seg_index]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return self._sig_timestamp0[seg_index] / self._sig_sampling_rate

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self._sig_length[seg_index]

        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"]
        (global_channel_indexes,) = np.nonzero(mask == stream_id)
        if channel_indexes is None:
            channel_indexes = slice(None)
        global_channel_indexes = global_channel_indexes[channel_indexes]

        if not self._gap_mode:
            sigs_chunk = np.zeros((i_stop - i_start, len(global_channel_indexes)), dtype="int16")
            # previous behavior block index are linear
            block_start = i_start // RECORD_SIZE
            block_stop = i_stop // RECORD_SIZE + 1
            sl0 = i_start % RECORD_SIZE
            sl1 = sl0 + (i_stop - i_start)

            for i, global_chan_index in enumerate(global_channel_indexes):
                data = self._sigs_memmap[seg_index][global_chan_index]
                sub = data[block_start:block_stop]
                sigs_chunk[:, i] = sub["samples"].flatten()[sl0:sl1]
        else:
            sigs_chunk = np.full(
                shape=(i_stop - i_start, len(global_channel_indexes)), fill_value=self.fill_gap_value, dtype="int16"
            )
            # slow mode
            for i, global_chan_index in enumerate(global_channel_indexes):
                data = self._sigs_memmap[seg_index][global_chan_index]
                timestamp0 = data[0]["timestamp"]

                # find first block
                block0 = np.searchsorted(data["timestamp"], timestamp0 + i_start, side="right") - 1
                block0_pos = data[block0]["timestamp"] - timestamp0

                if i_start - block0_pos > RECORD_SIZE:
                    # the block has gap!!
                    pos = -((i_start - block0_pos) % RECORD_SIZE)
                    block_index = block0 + 1
                else:
                    # the first block do not have gaps
                    shift0 = i_start - block0_pos

                    if shift0 + (i_stop - i_start) < RECORD_SIZE:
                        # protect when only one small block
                        pos = i_stop - i_start
                        sigs_chunk[:, i][:pos] = data[block0]["samples"][shift0 : shift0 + pos]
                    else:

                        pos = RECORD_SIZE - shift0
                        sigs_chunk[:, i][:pos] = data[block0]["samples"][shift0:]
                    block_index = block0 + 1

                # full block
                while block_index < data.size and data[block_index]["timestamp"] - timestamp0 < i_stop - RECORD_SIZE:
                    diff = data[block_index]["timestamp"] - data[block_index - 1]["timestamp"]
                    if diff > RECORD_SIZE:
                        # gap detected need jump
                        pos += diff - RECORD_SIZE

                    sigs_chunk[:, i][pos : pos + RECORD_SIZE] = data[block_index]["samples"][:]
                    pos += RECORD_SIZE
                    block_index += 1

                # last block
                if pos < i_stop - i_start:
                    diff = data[block_index]["timestamp"] - data[block_index - 1]["timestamp"]
                    if diff == RECORD_SIZE:
                        # ensure no gaps for last block
                        sigs_chunk[:, i][pos:] = data[block_index]["samples"][: i_stop - i_start - pos]

        return sigs_chunk

    def _get_spike_slice(self, seg_index, unit_index, t_start, t_stop):
        name, sorted_id = self.header["spike_channels"][unit_index]["name"].split("#")
        sorted_id = int(sorted_id)
        data_spike = self._spikes_memmap[seg_index][name]

        if t_start is None:
            t_start = self._first_spk_timestamps[seg_index]
        if t_stop is None:
            t_stop = self._last_spk_timestamps[seg_index]
        ts0 = int(t_start * self._spike_sampling_rate)
        ts1 = int(t_stop * self._spike_sampling_rate)

        ts = data_spike["timestamp"]
        keep = (data_spike["sorted_id"] == sorted_id) & (ts >= ts0) & (ts <= ts1)
        return data_spike, keep

    def _spike_count(self, block_index, seg_index, unit_index):
        data_spike, keep = self._get_spike_slice(seg_index, unit_index, None, None)
        return np.sum(keep)

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        data_spike, keep = self._get_spike_slice(seg_index, unit_index, t_start, t_stop)
        return data_spike["timestamp"][keep]

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype) / self._spike_sampling_rate
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        data_spike, keep = self._get_spike_slice(seg_index, unit_index, t_start, t_stop)
        nb_chan = data_spike[0]["nb_channel"]
        nb = np.sum(keep)
        waveforms = data_spike[keep]["samples"].flatten()
        waveforms = waveforms.reshape(nb, nb_chan, -1)
        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        # assert event_channel_index==0
        return self._events_memmap[seg_index].size

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        # assert event_channel_index==0

        if t_start is None:
            t_start = self._segment_t_start(block_index, seg_index)
        if t_stop is None:
            t_stop = self._segment_t_stop(block_index, seg_index)
        ts0 = int(t_start * self._event_sampling_rate)
        ts1 = int(t_stop * self._event_sampling_rate)
        ts = self._events_memmap[seg_index]["timestamp"]
        keep = (ts >= ts0) & (ts <= ts1)

        subdata = self._events_memmap[seg_index][keep]
        timestamps = subdata["timestamp"]
        # question what is the label????
        # here I put a combination
        labels = np.array(
            [f'{int(d["event_type"])}#{int(d["processor_id"])}#{int(d["chan_id"])}' for d in subdata],
            dtype="U",
        )
        durations = None

        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype) / self._event_sampling_rate
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        return None


continuous_dtype = [
    ("timestamp", "int64"),
    ("nb_sample", "uint16"),
    ("rec_num", "uint16"),
    ("samples", ">i2", RECORD_SIZE),
    ("markers", "uint8", 10),
]

events_dtype = [
    ("timestamp", "int64"),
    ("sample_pos", "int16"),
    ("event_type", "uint8"),
    ("processor_id", "uint8"),
    ("event_id", "uint8"),
    ("chan_id", "uint8"),
    ("record_num", "uint16"),
]

# the dtype is dynamic and depend on nb_channel and nb_sample
_base_spikes_dtype = [
    ("event_stype", "uint8"),
    ("timestamp", "int64"),
    ("software_timestamp", "int64"),
    ("source_id", "uint16"),
    ("nb_channel", "uint16"),
    ("nb_sample", "uint16"),
    ("sorted_id", "uint16"),
    ("electrode_id", "uint16"),
    ("within_chan_index", "uint16"),
    ("color", "uint8", 3),
    ("pca", "float32", 2),
    ("sampling_rate", "uint16"),
    ("samples", "uint16", None),
    ("gains", "float32", None),
    ("thresholds", "uint16", None),
    ("rec_num", "uint16"),
]


def make_spikes_dtype(filename):
    """
    Given the spike file make the appropriate dtype that depends on:
      * N - number of channels
      * M - samples per spike
    See documentation of file format.
    """

    # strangely the header do not have the sample size
    # So this do not work (too bad):
    # spike_info = read_file_header(filename)
    # N = spike_info['num_channels']
    # M =????

    # so we need to read the very first spike
    # but it will fail when 0 spikes (too bad)
    filesize = filename.stat().st_size
    if filesize >= (HEADER_SIZE + 23):
        with open(filename, mode="rb") as f:
            # M and N is at 1024 + 19 bytes
            f.seek(HEADER_SIZE + 19)
            N = np.fromfile(f, np.dtype("<u2"), 1)[0]
            M = np.fromfile(f, np.dtype("<u2"), 1)[0]
    else:
        spike_info = read_file_header(filename)
        N = spike_info["num_channels"]
        M = 40  # this is in the original code from openephys

    # make a copy
    spikes_dtype = [e for e in _base_spikes_dtype]
    spikes_dtype[12] = ("samples", "uint16", N * M)
    spikes_dtype[13] = ("gains", "float32", N)
    spikes_dtype[14] = ("thresholds", "uint16", N)

    return spikes_dtype


def explore_folder(dirname):
    """
    This explores a folder and dispatch continuous, event and spikes
    files by segment (aka recording session).

    The number of segments is checked with these rules
    "100_CH0.continuous" ---> seg_index 0
    "100_CH0_2.continuous" ---> seg_index 1
    "100_CH0_N.continuous" ---> seg_index N-1

    Newer formats follow similar rules but have an addition
    "100_RhythmData-A_CH0.continuous" ----> seg_index 0
    """
    filenames = [filename for filename in Path(dirname).glob("**/*") if filename.is_file()]
    filenames.sort()

    info = {}
    info["nb_segment"] = 0
    info["continuous"] = {}
    info["spikes"] = {}
    for filename in filenames:
        if filename.suffix == ".continuous":
            s = filename.stem.split("_")
            # For continuous files we check if the last value is an int indicating that a new segment should be
            # generated and if it is not an int then this must be same segment
            try:
                seg_index = int(s[-1]) - 1
            except ValueError:
                seg_index = 0
            if seg_index not in info["continuous"].keys():
                info["continuous"][seg_index] = []
            info["continuous"][seg_index].append(filename)
            if (seg_index + 1) > info["nb_segment"]:
                info["nb_segment"] += 1
        elif filename.suffix == ".spikes":
            s = re.findall(r"(_\d+)$", filename.stem)
            if s:
                seg_index = int(s[0][1:]) - 1
            else:
                seg_index = 0
            if seg_index not in info["spikes"].keys():
                info["spikes"][seg_index] = []
            info["spikes"][seg_index].append(filename)
            if (seg_index + 1) > info["nb_segment"]:
                info["nb_segment"] += 1

    # order continuous file by channel number within segment
    # order "CH before "ADC"
    for seg_index, continuous_filenames in info["continuous"].items():
        chan_ids_by_type = {}
        filenames_by_type = {}
        for continuous_filename in continuous_filenames:
            s = continuous_filename.stem.split("_")
            # new format includes putting a name between e.g. ['124', 'RhythmData', 'CH1']
            # old format would just be ['124', 'CH1']
            if len(s) == 2:
                processor_id, ch_name = s[0], s[1]
                chan_type = re.split(r"(\d+)", s[1])[0]
            else:
                processor_id, ch_name = s[0], s[2]
                chan_type = re.split(r"(\d+)", s[2])[0]
            chan_id = int(ch_name.replace(chan_type, ""))
            if chan_type in chan_ids_by_type.keys():
                chan_ids_by_type[chan_type].append(chan_id)
                filenames_by_type[chan_type].append(continuous_filename)
            else:
                chan_ids_by_type[chan_type] = [chan_id]
                filenames_by_type[chan_type] = [continuous_filename]
        chan_types = list(chan_ids_by_type.keys())

        if "CH" in chan_types:
            # force CH at beginning
            chan_types.remove("CH")
            chan_types = ["CH"] + chan_types

        ordered_continuous_filenames = []
        for chan_type in chan_types:
            local_order = np.argsort(chan_ids_by_type[chan_type])
            local_filenames = np.array(filenames_by_type[chan_type])[local_order]
            ordered_continuous_filenames.extend(local_filenames)
        info["continuous"][seg_index] = ordered_continuous_filenames

    # order spike files within segment
    for seg_index, spike_filenames in info["spikes"].items():
        names = []
        for spike_filename in spike_filenames:
            name = spike_filename.stem
            if seg_index > 0:
                name = name.replace("_" + str(seg_index + 1), "")
            names.append(name)
        order = np.argsort(names)
        spike_filenames = [spike_filenames[i] for i in order]
        info["spikes"][seg_index] = spike_filenames

    return info


def read_file_header(filename):
    """Read header information from the first 1024 bytes of an OpenEphys file.
    See docs.
    """
    header = {}
    with open(filename, mode="rb") as f:
        # Read the data as a string
        # Remove newlines and redundant "header." prefixes
        # The result should be a series of "key = value" strings, separated
        # by semicolons.
        header_string = f.read(HEADER_SIZE).replace(b"\n", b"").replace(b"header.", b"")

    # Parse each key = value string separately
    for pair in header_string.split(b";"):
        if b"=" in pair:
            key, value = pair.split(b" = ")
            key = key.strip().decode("ascii")
            value = value.strip()

            # Convert some values to numeric
            if key in ["bitVolts", "sampleRate"]:
                header[key] = float(value)
            elif key in ["blockLength", "bufferSize", "header_bytes", "num_channels"]:
                header[key] = int(value)
            else:
                # Keep as string
                header[key] = value.decode("ascii")

    return header



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.mearecrawio import MEArecRawIO


class MEArecIO(MEArecRawIO, BaseFromRaw):
    __doc__ = MEArecRawIO.__doc__
    mode = "file"

    def __init__(self, filename, load_spiketrains=True, load_analogsignal=True):
        MEArecRawIO.__init__(
            self, filename=filename, load_spiketrains=load_spiketrains, load_analogsignal=load_analogsignal
        )
        BaseFromRaw.__init__(self, filename)



"""
Tests of neo.io.neuroexplorerio
"""

import unittest

from neo.io import NeuroExplorerIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestNeuroExplorerIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = NeuroExplorerIO
    entities_to_download = ["neuroexplorer"]
    entities_to_download_test = [
        "neuroexplorer/File_neuroexplorer_1.nex",
        "neuroexplorer/File_neuroexplorer_2.nex",
    ]

    def test_signal_group_mode(self):
        filename = self.get_local_path("neuroexplorer/File_neuroexplorer_1.nex")

        # test that 2 signals are rendered with 2 sampling_rate
        for signal_group_mode in ("group-by-same-units", "split-all"):
            reader = NeuroExplorerIO(filename=filename)
            bl = reader.read_block(signal_group_mode=signal_group_mode)
            seg = bl.segments[0]
            assert len(seg.analogsignals) == 2
            anasig0 = seg.analogsignals[0]
            anasig1 = seg.analogsignals[1]
            assert anasig0.sampling_rate != anasig1.sampling_rate
            assert anasig0.shape != anasig1.shape
            # ~ for anasig in seg.analogsignals:
            # ~ print(anasig.shape, anasig.sampling_rate)


if __name__ == "__main__":
    unittest.main()



import numpy as np

from .baseio import BaseIO
from neo.core import ImageSequence, Segment, Block


class AsciiImageIO(BaseIO):
    """
    IO class for reading ImageSequence in a text file

    *Usage*:
        >>> from neo import io
        >>> import quantities as pq
        >>> r = io.AsciiImageIO(file_name='File_asciiimage_1.txt',nb_frame=511, nb_row=100,
        ...                     nb_column=100,units='mm', sampling_rate=1.0*pq.Hz,
        ...                     spatial_scale=1.0*pq.mm)
        >>> block = r.read_block()
        read block
        creating segment
        returning block
        >>> block
        Block with 1 segments
        file_origin: 'File_asciiimage_1.txt
        # segments (N=1)
        0: Segment with 1 imagesequences # analogsignals (N=0)

    """

    name = "AsciiImage IO"
    description = "Neo IO module for optical imaging data stored as a folder of TIFF images."

    _prefered_signal_group_mode = "group-by-same-units"
    is_readable = True
    is_writable = False

    supported_objects = [Block, Segment, ImageSequence]
    readable_objects = supported_objects
    writeable_object = []

    support_lazy = False

    read_params = {}
    write_params = {}

    extensions = []

    mode = "file"

    def __init__(
        self,
        file_name=None,
        nb_frame=None,
        nb_row=None,
        nb_column=None,
        units=None,
        sampling_rate=None,
        spatial_scale=None,
        **kwargs,
    ):

        BaseIO.__init__(self, file_name, **kwargs)
        self.nb_frame = nb_frame
        self.nb_row = nb_row
        self.nb_column = nb_column
        self.units = units
        self.sampling_rate = sampling_rate
        self.spatial_scale = spatial_scale

    def read_block(self, lazy=False, **kwargs):

        file = open(self.filename, "r")
        data = file.read()
        print("read block")
        liste_value = []
        record = []
        for i in range(len(data)):

            if data[i] == "\n" or data[i] == "\t":
                t = "".join(str(e) for e in record)
                liste_value.append(t)
                record = []
            else:
                record.append(data[i])

        data = []
        nb = 0
        for i in range(self.nb_frame):
            data.append([])
            for y in range(self.nb_row):
                data[i].append([])
                for x in range(self.nb_column):
                    data[i][y].append(liste_value[nb])
                    nb += 1

        image_sequence = ImageSequence(
            np.array(data, dtype="float"),
            units=self.units,
            sampling_rate=self.sampling_rate,
            spatial_scale=self.spatial_scale,
        )
        file.close()
        print("creating segment")
        segment = Segment(file_origin=self.filename)
        segment.imagesequences = [image_sequence]

        block = Block(file_origin=self.filename)
        block.segments.append(segment)
        print("returning block")

        return block



"""
Tests of neo.io.bci2000io
"""

import unittest

from neo.io import BCI2000IO
from neo.test.iotest.common_io_test import BaseTestIO


class TestBCI2000IO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = BCI2000IO
    entities_to_download = ["bci2000"]
    entities_to_test = [
        "bci2000/eeg1_1.dat",
        "bci2000/eeg1_2.dat",
        "bci2000/eeg1_3.dat",
    ]


if __name__ == "__main__":
    unittest.main()



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.spikeglxrawio import SpikeGLXRawIO


class SpikeGLXIO(SpikeGLXRawIO, BaseFromRaw):
    __doc__ = SpikeGLXRawIO.__doc__
    mode = "dir"

    def __init__(self, dirname, load_sync_channel=False, load_channel_location=False):
        SpikeGLXRawIO.__init__(
            self, dirname=dirname, load_sync_channel=load_sync_channel, load_channel_location=load_channel_location
        )
        BaseFromRaw.__init__(self, dirname)



"""
This generates a diagram in .png and .svg from neo.core objects


Authors: Samuel Garcia, Julia Sprenger, Andrew Davison
"""

from datetime import datetime

import numpy as np
import quantities as pq
from matplotlib import pyplot
from matplotlib.patches import Rectangle, ArrowStyle, FancyArrowPatch
from matplotlib.font_manager import FontProperties

import neo

line_height = 0.22
fontsize = 10.5
left_text_shift = 0.1
dpi = 100


def get_rect_height(name, obj):
    """
    calculate rectangle height
    """
    nlines = 1.5
    nlines += len(getattr(obj, "_all_attrs", []))
    nlines += len(getattr(obj, "_child_objects", []))
    if name == "Group":
        # because we represent RegionOfInterest only once
        # not the individual subclasses
        nlines -= 2
    return nlines * line_height


def annotate(ax, coord1, coord2, connectionstyle, color, alpha):
    arrowprops = dict(
        arrowstyle="simple",
        shrinkA=0.3,
        shrinkB=0.3,
        fc=color,
        ec=color,
        connectionstyle=connectionstyle,
        alpha=alpha,
    )
    bbox = dict(boxstyle="square", fc="w")
    a = ax.annotate(
        "",
        coord1,
        coord2,
        ha="right",
        va="center",
        size=fontsize,
        arrowprops=arrowprops,
        bbox=bbox,
    )
    a.set_zorder(-4)


def calc_coordinates(pos, height):
    x = pos[0]
    y = pos[1] + height - line_height * 0.5

    return x, y


def initialize_dummy_neo(name):
    if name in ["Block", "Segment", "Group", "Event", "Epoch"]:
        return getattr(neo, name)()
    elif name in ["ChannelView"]:
        sig = neo.AnalogSignal([] * pq.V, sampling_rate=1 * pq.Hz)
        return neo.ChannelView(sig, [0])
    elif name in ["ImageSequence"]:
        return neo.ImageSequence(
            np.array([1]).reshape((1, 1, 1)) * pq.V, spatial_scale=1 * pq.m, sampling_rate=1 * pq.Hz
        )
    elif name in ["AnalogSignal"]:
        return neo.AnalogSignal([] * pq.V, sampling_rate=1 * pq.Hz)
    elif name in ["SpikeTrain"]:
        return neo.SpikeTrain([] * pq.s, 0 * pq.s)
    elif name in ["IrregularlySampledSignal"]:
        return neo.IrregularlySampledSignal([] * pq.s, [] * pq.V)
    elif name in ["RegionOfInterest"]:
        imgseq = initialize_dummy_neo("ImageSequence")
        return neo.core.regionofinterest.CircularRegionOfInterest(imgseq, 0, 0, 0)
    else:
        raise ValueError(f"Unknown neo object: {name}")


def generate_diagram(rect_pos, rect_width, figsize):
    rw = rect_width

    fig = pyplot.figure(figsize=figsize)
    ax = fig.add_axes([0, 0, 1, 1])

    all_h = {}
    objs = {}
    for name in rect_pos:
        objs[name] = initialize_dummy_neo(name)
        all_h[name] = get_rect_height(name, objs[name])

    # draw connections
    color = ["c", "m", "r"]
    alpha = [0.8, 0.8, 0.5]
    CHILD_OBJECTS = 0
    CHILD_PROPERTIES = 1
    CUSTOM_RELATIONS = 2
    for name, pos in rect_pos.items():
        obj = objs[name]
        relationships = [getattr(obj, "_child_objects", []), getattr(obj, "_child_properties", [])]

        # additional references
        custom_relations = []
        for attr in obj._all_attrs:
            attr_name, attr_type = attr[0], attr[1]
            if isinstance(attr_type, str) and hasattr(neo, attr_type):
                custom_relations.append(attr_type)
            elif isinstance(attr_type, tuple):
                for attr_t in attr_type:
                    if isinstance(attr_t, str) and hasattr(neo, attr_t):
                        custom_relations.append(attr_t)
        relationships.append(custom_relations)

        for r in (CHILD_OBJECTS, CHILD_PROPERTIES, CUSTOM_RELATIONS):
            for child_name in relationships[r]:
                if ("RegionOfInterest") in child_name:
                    child_name = "RegionOfInterest"
                if child_name not in rect_pos:
                    continue
                x1, y1 = calc_coordinates(rect_pos[child_name], all_h[child_name])
                x2, y2 = calc_coordinates(pos, all_h[name])

                # autolink
                if child_name == name:
                    rad = 1
                    y1 += 0.2
                    x1 += 0.1
                else:
                    rad = (y1 - y2) / 50

                if r == CUSTOM_RELATIONS:
                    rad = -0.1

                if r in (CHILD_OBJECTS, CHILD_PROPERTIES, CUSTOM_RELATIONS):
                    x2 += rect_width
                connectionstyle = f"arc3,rad={rad}"

                annotate(
                    ax=ax,
                    coord1=(x1 - 0.02, y1 - 0.05),
                    coord2=(x2 - 0.05, y2),
                    connectionstyle=connectionstyle,
                    color=color[r],
                    alpha=alpha[r],
                )

    # draw boxes
    for name, pos in rect_pos.items():
        htotal = all_h[name]
        obj = objs[name]
        allrelationship = []
        for item in getattr(obj, "_child_containers", []):
            if item not in allrelationship:
                # this is to avoid having "regionsofinterest" appear three times in Group
                # for the three subclasses
                allrelationship.append(item)

        rect = Rectangle(pos, rect_width, htotal, facecolor="w", edgecolor="k", linewidth=2.0)
        ax.add_patch(rect)

        # title green
        pos2 = pos[0], pos[1] + htotal - line_height * 1.5
        rect = Rectangle(pos2, rect_width, line_height * 1.5, facecolor="g", edgecolor="k", alpha=0.5, linewidth=2.0)
        ax.add_patch(rect)

        # single relationship
        relationship = list(getattr(obj, "_child_objects", []))
        if name == "Group":
            relationship.remove("RectangularRegionOfInterest")
            relationship.remove("PolygonRegionOfInterest")
        pos2 = pos[1] + htotal - line_height * (1.5 + len(relationship))
        rect_height = len(relationship) * line_height

        rect = Rectangle((pos[0], pos2), rect_width, rect_height, facecolor="c", edgecolor="k", alpha=0.5)
        ax.add_patch(rect)

        # necessary attr
        pos2 = pos[1] + htotal - line_height * (1.5 + len(allrelationship) + len(obj._necessary_attrs))
        rect = Rectangle(
            (pos[0], pos2), rect_width, line_height * len(obj._necessary_attrs), facecolor="r", edgecolor="k", alpha=0.5
        )
        ax.add_patch(rect)

        # name
        if hasattr(obj, "_quantity_attr"):
            post = "* "
        else:
            post = ""
        ax.text(
            pos[0] + rect_width / 2.0,
            pos[1] + htotal - line_height * 1.5 / 2.0,
            name + post,
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=fontsize + 2,
            fontproperties=FontProperties(weight="bold"),
        )

        # relationship
        for i, relat in enumerate(allrelationship):
            ax.text(
                pos[0] + left_text_shift,
                pos[1] + htotal - line_height * (i + 2),
                relat + ": list",
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=fontsize,
            )
        # attributes
        for i, attr in enumerate(obj._all_attrs):
            attrname, attrtype = attr[0], attr[1]
            if hasattr(obj, "_quantity_attr") and obj._quantity_attr == attrname:
                t1 = attrname + " (object itself)"
            else:
                t1 = attrname

            if attrtype == pq.Quantity:
                if attr[2] == 0:
                    t2 = "Quantity scalar"
                else:
                    t2 = "Quantity %dD" % attr[2]
            elif attrtype == np.ndarray:
                t2 = "np.ndarray %dD dt='%s'" % (attr[2], attr[3].kind)
            elif attrtype == datetime:
                t2 = "datetime"
            elif type(attrtype) == tuple:
                t2 = str(attrtype).replace("(", "").replace(")", "")
                t2 = t2.replace("'", "").replace(", ", " or ")
            else:
                t2 = attrtype.__name__

            t = t1 + " :  " + t2

            # abbreviating lines to match in rectangles
            char_limit = int(rect_width * 13.4)  # 13.4 = arbitrary calibration
            if len(t) > char_limit:
                t = t[: char_limit - 3] + "..."
            ax.text(
                pos[0] + left_text_shift,
                pos[1] + htotal - line_height * (i + len(allrelationship) + 2),
                t,
                horizontalalignment="left",
                verticalalignment="center",
                fontsize=fontsize,
            )

    xlim, ylim = figsize
    ax.set_xlim(0, xlim)
    ax.set_ylim(0, ylim)

    ax.set_xticks([])
    ax.set_yticks([])

    return fig


def generate_diagram_simple():
    figsize = (18, 12)
    rw = rect_width = 3.0
    bf = 1.5
    rect_pos = {
        #  col 0
        "Block": (0.5 + rw * bf * 0, 7),
        #  col 1
        "Segment": (0.5 + rw * bf * 0.9, 8.5),
        "Group": (0.5 + rw * bf * 0.9, 3.7),
        #  col 2
        "ChannelView": (0.5 + rw * bf * 1.8, 2.5),
        "RegionOfInterest": (0.5 + rw * bf * 1.8, 1.0),
        # col 3
        "AnalogSignal": (0.5 + rw * bf * 3, 10),
        "IrregularlySampledSignal": (0.5 + rw * bf * 3, 8.4),
        "ImageSequence": (0.5 + rw * bf * 3, 6.35),
        "SpikeTrain": (0.5 + rw * bf * 3, 3.8),
        "Event": (0.5 + rw * bf * 3, 2.1),
        "Epoch": (0.5 + rw * bf * 3, 0.3),
    }

    fig = generate_diagram(rect_pos, rect_width, figsize)
    fig.savefig("simple_generated_diagram.png", dpi=dpi)
    fig.savefig("simple_generated_diagram.svg", dpi=dpi)


if __name__ == "__main__":
    generate_diagram_simple()
    pyplot.show()



"""
Tests proxyobject mechanisms with ExampleRawIO
"""

import unittest

import numpy as np
import quantities as pq
from neo.core import NeoReadWriteError
from neo.rawio.examplerawio import ExampleRawIO
from neo.io.proxyobjects import AnalogSignalProxy, SpikeTrainProxy, EventProxy, EpochProxy

from neo.core import Segment, AnalogSignal, Epoch, Event, SpikeTrain


from neo.test.tools import assert_arrays_almost_equal, assert_neo_object_is_compliant, assert_same_attributes


class BaseProxyTest(unittest.TestCase):
    def setUp(self):
        self.reader = ExampleRawIO(filename="my_filename.fake")
        self.reader.parse_header()


class TestAnalogSignalProxy(BaseProxyTest):

    def test_AnalogSignalProxy(self):
        proxy_anasig = AnalogSignalProxy(
            rawio=self.reader,
            stream_index=0,
            inner_stream_channels=None,
            block_index=0,
            seg_index=0,
        )

        assert proxy_anasig.sampling_rate == 10 * pq.kHz
        assert proxy_anasig.t_start == 0 * pq.s
        assert proxy_anasig.t_stop == 10 * pq.s
        assert proxy_anasig.duration == 10 * pq.s
        assert proxy_anasig.file_origin == "my_filename.fake"

        # full load
        full_anasig = proxy_anasig.load(time_slice=None)
        assert isinstance(full_anasig, AnalogSignal)
        assert_same_attributes(proxy_anasig, full_anasig)

        # slice time
        anasig = proxy_anasig.load(time_slice=(2.0 * pq.s, 5 * pq.s))
        assert anasig.t_start == 2.0 * pq.s
        assert anasig.duration == 3.0 * pq.s
        assert anasig.shape == (30000, 8)
        assert_same_attributes(proxy_anasig.time_slice(2.0 * pq.s, 5 * pq.s), anasig)

        # round up next sample when slicing
        anasig = proxy_anasig.load(time_slice=(1.99999 * pq.s, 5.000001 * pq.s))
        assert anasig.t_start == 2.0 * pq.s
        assert anasig.duration == 3.0 * pq.s
        assert anasig.shape == (30000, 8)

        # round down sample test
        anasig = proxy_anasig.load(time_slice=(2.00001 * pq.s, 4.99999 * pq.s))
        assert anasig.t_start == 2.0 * pq.s
        assert anasig.duration == 3.0 * pq.s
        assert anasig.shape == (30000, 8)

        # buggy time slice
        with self.assertRaises(ValueError):
            anasig = proxy_anasig.load(time_slice=(2.0 * pq.s, 15 * pq.s))
        anasig = proxy_anasig.load(time_slice=(2.0 * pq.s, 15 * pq.s), strict_slicing=False)
        assert proxy_anasig.t_stop == 10 * pq.s

        # select channels
        anasig = proxy_anasig.load(channel_indexes=[3, 4, 5])
        assert anasig.shape[1] == 3

        # select channels and slice times
        anasig = proxy_anasig.load(time_slice=(2.0 * pq.s, 5 * pq.s), channel_indexes=[3, 4, 5])
        assert anasig.shape == (30000, 3)

        # magnitude mode rescaled
        anasig_float = proxy_anasig.load(magnitude_mode="rescaled")
        assert anasig_float.dtype == "float32"
        assert anasig_float.units == pq.uV
        assert anasig_float.units == proxy_anasig.units

        # magnitude mode raw
        anasig_int = proxy_anasig.load(magnitude_mode="raw")
        assert anasig_int.dtype == "int16"
        assert anasig_int.units == pq.CompoundUnit("0.0152587890625*uV")

        assert_arrays_almost_equal(anasig_float, anasig_int.rescale("uV"), 1e-9)

        # test array_annotations
        assert "__array_annotations__" not in proxy_anasig.annotations
        assert "impedance" in proxy_anasig.array_annotations
        assert proxy_anasig.array_annotations["impedance"].size == 8
        assert "impedance" in anasig_float.array_annotations
        assert anasig_float.array_annotations["impedance"].size == 8

    def test_global_local_channel_indexes(self):
        proxy_anasig = AnalogSignalProxy(
            rawio=self.reader, stream_index=0, inner_stream_channels=slice(0, 8, 2), block_index=0, seg_index=0
        )

        assert proxy_anasig.shape == (100000, 4)
        assert np.array_equal(proxy_anasig.array_annotations["channel_names"], ["ch0", "ch2", "ch4", "ch6"])

        # should be channel ch0 and ch6
        anasig = proxy_anasig.load(channel_indexes=[0, 3])
        assert anasig.shape == (100000, 2)
        assert np.array_equal(anasig.array_annotations["channel_names"], ["ch0", "ch6"])


class TestSpikeTrainProxy(BaseProxyTest):
    def test_SpikeTrainProxy(self):
        proxy_sptr = SpikeTrainProxy(rawio=self.reader, spike_channel_index=0, block_index=0, seg_index=0)

        assert proxy_sptr.name == "unit0"
        assert proxy_sptr.t_start == 0 * pq.s
        assert proxy_sptr.t_stop == 10 * pq.s
        assert proxy_sptr.shape == (20,)
        assert proxy_sptr.left_sweep == 0.002 * pq.s
        assert proxy_sptr.sampling_rate == 10 * pq.kHz

        # full load
        full_sptr = proxy_sptr.load(time_slice=None)
        assert isinstance(full_sptr, SpikeTrain)
        assert_same_attributes(proxy_sptr, full_sptr)
        assert full_sptr.shape == proxy_sptr.shape

        # slice time
        sptr = proxy_sptr.load(time_slice=(250 * pq.ms, 500 * pq.ms))
        assert sptr.t_start == 0.25 * pq.s
        assert sptr.t_stop == 0.5 * pq.s
        assert sptr.shape == (6,)
        assert_same_attributes(proxy_sptr.time_slice(250 * pq.ms, 500 * pq.ms), sptr)

        # buggy time slice
        with self.assertRaises(ValueError):
            sptr = proxy_sptr.load(time_slice=(2.0 * pq.s, 15 * pq.s))
        sptr = proxy_sptr.load(time_slice=(2.0 * pq.s, 15 * pq.s), strict_slicing=False)
        assert sptr.t_stop == 10 * pq.s

        # magnitude mode rescaled
        sptr_float = proxy_sptr.load(magnitude_mode="rescaled")
        assert sptr_float.dtype == "float64"
        assert sptr_float.units == pq.s

        # magnitude mode raw
        # TODO when raw mode implemented
        # sptr_int = proxy_sptr.load(magnitude_mode='raw')
        # assert sptr_int.dtype=='int64'
        # assert sptr_int.units==pq.CompoundUnit('1/10000*s')

        # assert_arrays_almost_equal(sptr_float, sptr_int.rescale('s'), 1e-9)

        # Without waveforms
        sptr = proxy_sptr.load(load_waveforms=False)
        assert sptr.waveforms is None

        # With waveforms
        sptr = proxy_sptr.load(load_waveforms=True, magnitude_mode="rescaled")
        assert sptr.waveforms is not None
        assert sptr.waveforms.shape == (20, 1, 50)
        assert sptr.waveforms.units == 1 * pq.uV

        # slice waveforms
        sptr = proxy_sptr.load(load_waveforms=True, time_slice=(250 * pq.ms, 500 * pq.ms))
        assert sptr.waveforms.shape == (6, 1, 50)

        # test array_annotations
        assert "__array_annotations__" not in proxy_sptr.annotations
        assert "amplitudes" in proxy_sptr.array_annotations


class TestEventProxy(BaseProxyTest):
    def test_EventProxy(self):
        proxy_event = EventProxy(rawio=self.reader, event_channel_index=0, block_index=0, seg_index=0)

        assert proxy_event.name == "Some events"
        assert proxy_event.shape == (6,)

        # full load
        full_event = proxy_event.load(time_slice=None)
        assert isinstance(full_event, Event)
        assert_same_attributes(proxy_event, full_event, exclude=("times", "labels"))
        assert full_event.shape == proxy_event.shape

        # slice time
        event = proxy_event.load(time_slice=(1 * pq.s, 2 * pq.s))
        assert event.shape == (2,)
        assert event.labels.shape == (2,)
        assert_same_attributes(proxy_event.time_slice(1 * pq.s, 2 * pq.s), event)

        # buggy time slice
        with self.assertRaises(ValueError):
            event = proxy_event.load(time_slice=(2 * pq.s, 15 * pq.s))
        event = proxy_event.load(time_slice=(2 * pq.s, 15 * pq.s), strict_slicing=False)

        # test annotations/array_annotations
        assert "__array_annotations__" not in proxy_event.annotations
        assert "nickname" in proxy_event.annotations
        assert "button" in proxy_event.array_annotations


class TestEpochProxy(BaseProxyTest):
    def test_EpochProxy(self):
        proxy_epoch = EpochProxy(rawio=self.reader, event_channel_index=1, block_index=0, seg_index=0)

        assert proxy_epoch.name == "Some epochs"
        assert proxy_epoch.shape == (10,)

        # full load
        full_epoch = proxy_epoch.load(time_slice=None)
        assert isinstance(full_epoch, Epoch)
        assert_same_attributes(proxy_epoch, full_epoch, exclude=("times", "labels", "durations"))
        assert full_epoch.shape == proxy_epoch.shape

        # slice time
        epoch = proxy_epoch.load(time_slice=(1 * pq.s, 4 * pq.s))
        assert epoch.shape == (3,)
        assert epoch.labels.shape == (3,)
        assert epoch.durations.shape == (3,)
        assert_same_attributes(proxy_epoch.time_slice(1 * pq.s, 4 * pq.s), epoch)

        # buggy time slice
        with self.assertRaises(ValueError):
            epoch = proxy_epoch.load(time_slice=(2 * pq.s, 15 * pq.s))
        epoch = proxy_epoch.load(time_slice=(2 * pq.s, 15 * pq.s), strict_slicing=False)

        # test annotations/array_annotations
        assert "__array_annotations__" not in proxy_epoch.annotations
        assert "nickname" in proxy_epoch.annotations


class TestSegmentWithProxy(BaseProxyTest):
    def test_segment_with_proxy(self):
        seg = Segment()

        proxy_anasig = AnalogSignalProxy(
            rawio=self.reader,
            stream_index=0,
            inner_stream_channels=None,
            block_index=0,
            seg_index=0,
        )
        seg.analogsignals.append(proxy_anasig)

        proxy_sptr = SpikeTrainProxy(rawio=self.reader, spike_channel_index=0, block_index=0, seg_index=0)
        seg.spiketrains.append(proxy_sptr)

        proxy_event = EventProxy(rawio=self.reader, event_channel_index=0, block_index=0, seg_index=0)
        seg.events.append(proxy_event)

        proxy_epoch = EpochProxy(rawio=self.reader, event_channel_index=1, block_index=0, seg_index=0)
        seg.epochs.append(proxy_epoch)


if __name__ == "__main__":
    unittest.main()



"""
Class for reading data from CED (Cambridge Electronic Design)
http://ced.co.uk/

This read *.smrx (and *.smr) from spike2 and signal software.

Note Spike2RawIO/Spike2IO is the old implementation in neo.
It still works without any dependency and should be faster.
Spike2IO only works for smr (32 bit) and not for smrx (64 bit) files.

This implementation depends on the SONPY package:
https://pypi.org/project/sonpy/

Please note that the SONPY package:
  * is NOT open source
  * internally uses a list instead of numpy.ndarray, potentially causing slow data reading
  *  is maintained by CED


Author : Samuel Garcia
"""

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class CedRawIO(BaseRawIO):
    """
    Class for reading data from CED (Cambridge Electronic Design) spike2.

    Parameters
    ----------
    filename: str, default: ''
        The *.smr or *.smrx file to load
    take_ideal_sampling_rate: bool, default: False
        If true use the `GetIdealRate` function from sonpy package

    Notes
    -----
    * This internally uses the sonpy package which is closed source.

    * This IO reads smr and smrx files

    """

    extensions = ["smr", "smrx"]
    rawmode = "one-file"

    def __init__(
        self,
        filename="",
        take_ideal_sampling_rate=False,
    ):
        BaseRawIO.__init__(self)
        self.filename = filename

        self.take_ideal_sampling_rate = take_ideal_sampling_rate

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        import sonpy

        self.smrx_file = sonpy.lib.SonFile(sName=str(self.filename), bReadOnly=True)
        smrx = self.smrx_file

        self._time_base = smrx.GetTimeBase()

        channel_infos = []
        signal_channels = []
        spike_channels = []
        self._all_spike_ticks = {}

        for chan_ind in range(smrx.MaxChannels()):
            chan_type = smrx.ChannelType(chan_ind)
            chan_id = str(chan_ind)
            if chan_type == sonpy.lib.DataType.Adc:
                physical_chan = smrx.PhysicalChannel(chan_ind)
                divide = smrx.ChannelDivide(chan_ind)
                if self.take_ideal_sampling_rate:
                    sr = smrx.GetIdealRate(chan_ind)
                else:
                    sr = 1.0 / (smrx.GetTimeBase() * divide)
                max_time = smrx.ChannelMaxTime(chan_ind)
                first_time = smrx.FirstTime(chan_ind, 0, max_time)
                # max_times is included so +1
                time_size = (max_time - first_time) / divide + 1
                channel_infos.append((first_time, max_time, divide, time_size, sr))
                gain = smrx.GetChannelScale(chan_ind) / 6553.6
                offset = smrx.GetChannelOffset(chan_ind)
                units = smrx.GetChannelUnits(chan_ind)
                ch_name = smrx.GetChannelTitle(chan_ind)

                dtype = "int16"
                # set later after grouping
                stream_id = "0"
                signal_channels.append((ch_name, chan_id, sr, dtype, units, gain, offset, stream_id))

            elif chan_type == sonpy.lib.DataType.AdcMark:
                # spike and waveforms : only spike times is used here
                ch_name = smrx.GetChannelTitle(chan_ind)
                first_time = smrx.FirstTime(chan_ind, 0, max_time)
                max_time = smrx.ChannelMaxTime(chan_ind)
                divide = smrx.ChannelDivide(chan_ind)
                # here we don't use filter (sonpy.lib.MarkerFilter()) so we get all marker
                wave_marks = smrx.ReadWaveMarks(chan_ind, int(max_time / divide), 0, max_time)

                # here we load in memory all spike once because the access is really slow
                # with the ReadWaveMarks
                spike_ticks = np.array([t.Tick for t in wave_marks])
                spike_codes = np.array([t.Code1 for t in wave_marks])

                unit_ids = np.unique(spike_codes)
                for unit_id in unit_ids:
                    name = f"{ch_name}#{unit_id}"
                    spike_chan_id = f"ch{chan_id}#{unit_id}"
                    spike_channels.append((name, spike_chan_id, "", 1, 0, 0, 0))
                    mask = spike_codes == unit_id
                    self._all_spike_ticks[spike_chan_id] = spike_ticks[mask]

        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        # channels are grouped into stream if they have a common start, stop, size, divide and sampling_rate
        channel_infos = np.array(
            channel_infos,
            dtype=[("first_time", "i8"), ("max_time", "i8"), ("divide", "i8"), ("size", "i8"), ("sampling_rate", "f8")],
        )
        unique_info = np.unique(channel_infos)
        self.stream_info = unique_info
        signal_streams = []
        for i, info in enumerate(unique_info):
            stream_id = str(i)
            mask = channel_infos == info
            signal_channels["stream_id"][mask] = stream_id
            num_chans = np.sum(mask)
            stream_name = f"{stream_id} {num_chans}chans"
            signal_streams.append((stream_name, stream_id))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # spike channels not handled
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # event channels not handled
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self._seg_t_start = np.inf
        self._seg_t_stop = -np.inf
        for info in self.stream_info:
            self._seg_t_start = min(self._seg_t_start, info["first_time"] * self._time_base)

            self._seg_t_stop = max(self._seg_t_stop, info["max_time"] * self._time_base)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        return self._seg_t_start

    def _segment_t_stop(self, block_index, seg_index):
        return self._seg_t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        size = self.stream_info[stream_index]["size"]
        return size

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        info = self.stream_info[stream_index]
        t_start = info["first_time"] * self._time_base
        return t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):

        if i_start is None:
            i_start = 0
        if i_stop is None:
            i_stop = self.stream_info[stream_index]["size"]

        stream_id = self.header["signal_streams"]["id"][stream_index]
        signal_channels = self.header["signal_channels"]
        mask = signal_channels["stream_id"] == stream_id
        signal_channels = signal_channels[mask]
        if channel_indexes is not None:
            signal_channels = signal_channels[channel_indexes]

        num_chans = len(signal_channels)

        size = i_stop - i_start
        sigs = np.zeros((size, num_chans), dtype="int16")

        info = self.stream_info[stream_index]
        t_from = info["first_time"] + info["divide"] * i_start
        t_upto = info["first_time"] + info["divide"] * i_stop

        for i, chan_id in enumerate(signal_channels["id"]):
            chan_ind = int(chan_id)
            sig = self.smrx_file.ReadInts(chan=chan_ind, nMax=size, tFrom=t_from, tUpto=t_upto)
            sigs[:, i] = sig

        return sigs

    def _spike_count(self, block_index, seg_index, unit_index):
        unit_id = self.header["spike_channels"][unit_index]["id"]
        spike_ticks = self._all_spike_ticks[unit_id]
        return spike_ticks.size

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        unit_id = self.header["spike_channels"][unit_index]["id"]
        spike_ticks = self._all_spike_ticks[unit_id]
        if t_start is not None:
            tick_start = int(t_start / self._time_base)
            spike_ticks = spike_ticks[spike_ticks >= tick_start]
        if t_stop is not None:
            tick_stop = int(t_stop / self._time_base)
            spike_ticks = spike_ticks[spike_ticks <= tick_stop]
        return spike_ticks

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        spike_times = spike_timestamps.astype(dtype)
        spike_times *= self._time_base
        return spike_times

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        return None



"""
Class for reading data from Elan.

Elan is software for studying time-frequency maps of EEG data.

Elan is developed in Lyon, France, at INSERM U821

https://elan.lyon.inserm.fr

An Elan dataset is separated into 3 files :
 - .eeg          raw data file
 - .eeg.ent      header file
 - .eeg.pos      event file

Author: Samuel Garcia

"""

import datetime
import re
import pathlib

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)

from neo.core import NeoReadWriteError


class ElanRawIO(BaseRawIO):
    """
    Class for reading time-frequency EEG data maps from the Elan software

    Parameters
    ----------
    filename: str | None, default: None
        The raw data to load (*.eeg)
    entfile: str | None, default: None
        The header file to load (*.eeg.ent)
    posfile: str | None, deafult: None
        The event file to load (*.eeg.pos)

    Examples
    --------
    >>> import neo.rawio
    >>> reader = neo.rawio.ElanRawIO(filename='data.eeg', entfile='header.eeg.ent', posfile='events.eeg.pos')
    >>> reader.parse_header()
    >>> print(reader)
    """

    extensions = ["eeg"]
    rawmode = "one-file"

    def __init__(self, filename=None, entfile=None, posfile=None):
        BaseRawIO.__init__(self)
        self.filename = pathlib.Path(filename)

        # check whether ent and pos files are defined
        # keep existing suffixes in the process of ent and pos filename
        # generation
        if entfile is None:
            entfile = self.filename.with_suffix(self.filename.suffix + ".ent")
        if posfile is None:
            posfile = self.filename.with_suffix(self.filename.suffix + ".pos")
        self.entfile = entfile
        self.posfile = posfile

    def _parse_header(self):

        with open(self.entfile, mode="rt", encoding="ascii", newline=None) as f:

            # version
            version = f.readline()[:-1]
            if version not in ["V2", "V3"]:
                raise NeoReadWriteError(f"Reading is only possible for V2 or V3 .eeg.ent files. But {version=}")

            # info
            info1 = f.readline()[:-1]
            info2 = f.readline()[:-1]

            # strange 2 line for datetime
            # line1
            line = f.readline()
            r1 = re.findall(r"(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+)", line)
            r2 = re.findall(r"(\d+):(\d+):(\d+)", line)
            r3 = re.findall(r"(\d+)-(\d+)-(\d+)", line)
            YY, MM, DD, hh, mm, ss = (None,) * 6
            if len(r1) != 0:
                DD, MM, YY, hh, mm, ss = r1[0]
            elif len(r2) != 0:
                hh, mm, ss = r2[0]
            elif len(r3) != 0:
                DD, MM, YY = r3[0]

            # line2
            line = f.readline()
            r1 = re.findall(r"(\d+)-(\d+)-(\d+) (\d+):(\d+):(\d+)", line)
            r2 = re.findall(r"(\d+):(\d+):(\d+)", line)
            r3 = re.findall(r"(\d+)-(\d+)-(\d+)", line)
            if len(r1) != 0:
                DD, MM, YY, hh, mm, ss = r1[0]
            elif len(r2) != 0:
                hh, mm, ss = r2[0]
            elif len(r3) != 0:
                DD, MM, YY = r3[0]
            try:
                fulldatetime = datetime.datetime(int(YY), int(MM), int(DD), int(hh), int(mm), int(ss))
            except:
                fulldatetime = None

            line = f.readline()
            line = f.readline()
            line = f.readline()

            # sampling rate sample
            line = f.readline()
            self._sampling_rate = 1.0 / float(line)

            # nb channel
            line = f.readline()
            nb_channel = int(line) - 2

            channel_infos = [{} for c in range(nb_channel + 2)]
            # channel label
            for c in range(nb_channel + 2):
                channel_infos[c]["label"] = f.readline()[:-1]
            # channel kind
            for c in range(nb_channel + 2):
                channel_infos[c]["kind"] = f.readline()[:-1]
            # channel unit
            for c in range(nb_channel + 2):
                channel_infos[c]["units"] = f.readline()[:-1]
            # range for gain and offset
            for c in range(nb_channel + 2):
                channel_infos[c]["min_physic"] = float(f.readline()[:-1])
            for c in range(nb_channel + 2):
                channel_infos[c]["max_physic"] = float(f.readline()[:-1])
            for c in range(nb_channel + 2):
                channel_infos[c]["min_logic"] = float(f.readline()[:-1])
            for c in range(nb_channel + 2):
                channel_infos[c]["max_logic"] = float(f.readline()[:-1])

            # info filter
            info_filter = []
            for c in range(nb_channel + 2):
                channel_infos[c]["info_filter"] = f.readline()[:-1]

        n = int(round(np.log(channel_infos[0]["max_logic"] - channel_infos[0]["min_logic"]) / np.log(2)) / 8)
        sig_dtype = np.dtype(">i" + str(n))

        signal_streams = np.array([("Signals", "0")], dtype=_signal_stream_dtype)

        sig_channels = []
        for c, chan_info in enumerate(channel_infos[:-2]):
            chan_name = chan_info["label"]
            chan_id = str(c)

            gain = (chan_info["max_physic"] - chan_info["min_physic"]) / (
                chan_info["max_logic"] - chan_info["min_logic"]
            )
            offset = -chan_info["min_logic"] * gain + chan_info["min_physic"]
            stream_id = "0"
            sig_channels.append(
                (chan_name, chan_id, self._sampling_rate, sig_dtype, chan_info["units"], gain, offset, stream_id)
            )

        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # raw data
        self._raw_signals = np.memmap(self.filename, dtype=sig_dtype, mode="r", offset=0).reshape(-1, nb_channel + 2)
        self._raw_signals = self._raw_signals[:, :-2]

        # triggers
        with open(self.posfile, mode="rt", encoding="ascii", newline=None) as f:
            self._raw_event_timestamps = []
            self._event_labels = []
            self._reject_codes = []
            for line in f.readlines():
                r = re.findall(r" *(\d+)\s* *(\d+)\s* *(\d+) *", line)
                self._raw_event_timestamps.append(int(r[0][0]))
                self._event_labels.append(str(r[0][1]))
                self._reject_codes.append(str(r[0][2]))

        self._raw_event_timestamps = np.array(self._raw_event_timestamps, dtype="int64")
        self._event_labels = np.array(self._event_labels, dtype="U")
        self._reject_codes = np.array(self._reject_codes, dtype="U")

        event_channels = []
        event_channels.append(("Trigger", "", "event"))
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()
        extra_info = dict(rec_datetime=fulldatetime, elan_version=version, info1=info1, info2=info2)
        block_annotations = self.raw_annotations["blocks"][0]
        block_annotations.update(extra_info)
        seg_annotations = self.raw_annotations["blocks"][0]["segments"][0]
        seg_annotations.update(extra_info)

        sig_annotations = self.raw_annotations["blocks"][0]["segments"][0]["signals"][0]
        for key in ("info_filter", "kind"):
            values = [channel_infos[c][key] for c in range(nb_channel)]
            sig_annotations["__array_annotations__"][key] = np.array(values)

        event_annotations = self.raw_annotations["blocks"][0]["segments"][0]["events"][0]
        event_annotations["__array_annotations__"]["reject_codes"] = self._reject_codes

    def _source_name(self):
        return self.filename

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._raw_signals.shape[0] / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._raw_signals.shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(None)
        raw_signals = self._raw_signals[slice(i_start, i_stop), channel_indexes]
        return raw_signals

    def _spike_count(self, block_index, seg_index, unit_index):
        return 0

    def _event_count(self, block_index, seg_index, event_channel_index):
        return self._raw_event_timestamps.size

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        timestamp = self._raw_event_timestamps
        labels = self._event_labels
        durations = None

        if t_start is not None:
            keep = timestamp >= int(t_start * self._sampling_rate)
            timestamp = timestamp[keep]
            labels = labels[keep]

        if t_stop is not None:
            keep = timestamp <= int(t_stop * self._sampling_rate)
            timestamp = timestamp[keep]
            labels = labels[keep]

        return timestamp, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_times = event_timestamps.astype(dtype) / self._sampling_rate
        return event_times



"""
This module implements :class:`FilterCondition`, which enables use of different filter conditions
for neo.core.container.filter.

Classes:
    - :class:`FilterCondition`: Abstract base class for defining filter conditions.
    - :class:`Equals`: Filter condition to check if a value is equal to the control value.
    - :class:`IsNot`: Filter condition to check if a value is not equal to the control value.
    - :class:`LessThanOrEquals`: Filter condition to check if a value is less than or equal to the
    control value.
    - :class:`GreaterThanOrEquals`: Filter condition to check if a value is greater than or equal to
    the control value.
    - :class:`LessThan`: Filter condition to check if a value is less than the control value.
    - :class:`GreaterThan`: Filter condition to check if a value is greater than the control value.
    - :class:`IsIn`: Filter condition to check if a value is in a list or equal to the control
    value.
    - :class:`InRange`: Filter condition to check if a value is in a specified range.

The provided classes allow users to select filter conditions and use them with
:func:`neo.core.container.filter()` to perform specific filtering operations on data.
"""

from abc import ABC, abstractmethod
from numbers import Number
from typing import Union, Any


class FilterCondition(ABC):
    """
    FilterCondition object is given as parameter to container.filter():

    Usage:
        segment.filter(my_annotation=<FilterCondition>) or
        segment=filter({'my_annotation': <FilterCondition>})
    """

    @abstractmethod
    def __init__(self, control: Any) -> None:
        """
        Initialize new FilterCondition object.

        Parameters
        ----------
        control: Any
            The control value to be used for filtering.

        Notes
        -----
        This is an abstract base class and should not be instantiated directly.
        """

    @abstractmethod
    def evaluate(self, compare: Any) -> bool:
        """
        Evaluate the filter condition for given value.

        Parameters:
        -----------
        compare: Any
        The value to be compared with the control value.

        Returns
        -------
        bool: True if the condition is satisfied, False otherwise.

        Notes
        -----
        This method should be implemented in subclasses.
        """


class Equals(FilterCondition):
    """
    Filter condition to check if target value is equal to the control value.
    """

    def __init__(self, control: Any) -> None:
        self.control = control

    def evaluate(self, compare: Any) -> bool:
        return compare == self.control


class IsNot(FilterCondition):
    """
    Filter condition to check if target value is not equal to the control value.
    """

    def __init__(self, control: Any) -> None:
        self.control = control

    def evaluate(self, compare: Any) -> bool:
        return compare != self.control


class LessThanOrEquals(FilterCondition):
    """
    Filter condition to check if target value is less than or equal to the control value.
    """

    def __init__(self, control: Number) -> None:
        self.control = control

    def evaluate(self, compare: Number) -> bool:
        return compare <= self.control


class GreaterThanOrEquals(FilterCondition):
    """
    Filter condition to check if target value is greater than or equal to the control value.
    """

    def __init__(self, control: Number) -> None:
        self.control = control

    def evaluate(self, compare: Number) -> bool:
        return compare >= self.control


class LessThan(FilterCondition):
    """
    Filter condition to check if target value is less than the control value.
    """

    def __init__(self, control: Number) -> None:
        self.control = control

    def evaluate(self, compare: Number) -> bool:
        return compare < self.control


class GreaterThan(FilterCondition):
    """
    Filter condition to check if target value is greater than the control value.
    """

    def __init__(self, control: Number) -> None:
        self.control = control

    def evaluate(self, compare: Number) -> bool:
        return compare > self.control


class IsIn(FilterCondition):
    """
    Filter condition to check if target is in control.
    """

    def __init__(self, control: Union[list, tuple, set, int]) -> None:
        self.control = control

    def evaluate(self, compare: Any) -> bool:
        if isinstance(self.control, (list, tuple, set)):
            return compare in self.control
        if isinstance(self.control, int):
            return compare == self.control

        raise SyntaxError("parameter not of type list, tuple, set or int")


class InRange(FilterCondition):
    """
    Filter condition to check if a value is in a specified range.

    Parameters
    -----------
    lower_bound: int
        The lower bound of the range.
    upper_bound: int
        The upper bound of the range.
    left_closed: bool
        If True, the range includes the lower bound (lower_bound <= compare).
    right_closed: bool
        If True, the range includes the upper bound (compare <= upper_bound).

    Returns
    -------
    bool:
    whether the values are in range

    Examples
    --------

    >>> InRange(lower_bound, upper_bound, left_closed=False, right_closed=False)
    """

    def __init__(
        self, lower_bound: Number, upper_bound: Number, left_closed: bool = False, right_closed: bool = False
    ) -> None:
        if not isinstance(lower_bound, Number) or not isinstance(upper_bound, Number):
            raise ValueError("parameter is not a number")

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.left_closed = left_closed
        self.right_closed = right_closed

    def evaluate(self, compare: Number) -> bool:
        if not self.left_closed and not self.right_closed:
            return self.lower_bound <= compare <= self.upper_bound
        if not self.left_closed and self.right_closed:
            return self.lower_bound <= compare < self.upper_bound
        if self.left_closed and not self.right_closed:
            return self.lower_bound < compare <= self.upper_bound
        return self.lower_bound < compare < self.upper_bound



"""
Common tools that are useful for neo.io object tests
"""

import logging
import os

logger = logging.getLogger("neo.test")


def can_use_network():
    """
    Return True if network access is allowed
    """

    if os.environ.get("NEO_TESTS_NO_NETWORK", False):
        return False
    try:
        import datalad

        HAVE_DATALAD = True
    except:
        HAVE_DATALAD = False
    if not HAVE_DATALAD:
        return False
    return True
